#########
# Author:        gq1
# Maintainer:    $Author: gq1 $
# Created:       2009-04-17
# Last Modified: $Date: 2011-03-17 14:00:00 +0000 (Thu, 17 Mar 2011) $
# Id:            $Id: Alignment.pm 12834 2011-03-17 14:00:00Z gq1 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-43.0/lib/npg_common/Alignment.pm $
#

package npg_common::Alignment;

use strict;
use warnings;
use Carp;
use English qw(-no_match_vars);
use base qw(Class::Accessor);
use File::Temp qw(tempfile tempdir);
use IPC::Open3;
use Symbol;


use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 12834 $ =~ /(\d+)/mxs; $r; };

use Parallel::ForkManager;

__PACKAGE__->mk_accessors(qw(bwa_cmd bwa_options samtools_cmd sam_original_quality_cmd sam_original_quality_options sam_second_call_cmd sam_second_call_options sam_index_tag_cmd sam_header_cmd picard_converter_cmd sam_fastq_check_cmd));

sub new {
  my ($class, $ref) = @_;
  $ref ||= {};
  bless $ref, $class;

  return $ref;
}

#given one fastq and one reference, generate bam file for single end
sub bwa_align_bam_se {
  my ($self, $args_ref) = @_;

  my $query_fastq   = $args_ref->{fastq};
  my $bam_out       = $args_ref->{bam_out};
  my $ref_root      = $args_ref->{ref_root};
  my $ref_fai       = $args_ref->{ref_fai};

  my $work_dir = tempdir(CLEANUP => 1);
  my $sai_out = $work_dir.q{/sai};

  #alignment
  my $bwa_aln_cmd = $self->bwa_align( $ref_root, $query_fastq, $sai_out );

  $args_ref->{sai} = $sai_out;
  $args_ref->{bwa_aln_cmd} = $bwa_aln_cmd;

  #sai to bam
  $self->sai2bam_se( $args_ref );

  #remove sai file when sam file generated to save disk space
  unlink $sai_out;

  return 1;
}

#given two fastq and one reference, generate sam file for paired ends
sub bwa_align_bam_pe {
  my ( $self, $args_ref ) = @_;

  my $ref_root   = $args_ref->{ref_root};
  my $fastq1     = $args_ref->{fastq1};
  my $fastq2     = $args_ref->{fastq2};
  my $bam_out    = $args_ref->{bam_out};
  my $fork_align = $args_ref->{fork_align};

  my $work_dir = tempdir(CLEANUP => 1);

  my $sai_out_1 = $work_dir.q{/1.sai};
  my $sai_out_2 = $work_dir.q{/2.sai};

  #alignments
  my $bwa_aln_cmd = $self->bwa_align_list( $ref_root, [$fastq1, $fastq2], [$sai_out_1, $sai_out_2], $fork_align);

  $args_ref->{sai_1} = $sai_out_1;
  $args_ref->{sai_2} = $sai_out_2;
  $args_ref->{bwa_aln_cmd} = $bwa_aln_cmd;

  #sai to bam
  $self->sai2bam_pe( $args_ref );

  #remove sai files
  unlink $sai_out_1;
  unlink $sai_out_2;

  return 1;
}


#given one fastq and one reference, generate bam file for single end
sub bwa_align_se {
  my ($self,$args_ref) = @_;

  my $query_fastq  = $args_ref->{fastq};
  my $bam_out  = $args_ref->{bam_out};
  my $ref_root  = $args_ref->{ref_root};

  my $hit_sequence_ids     = $args_ref->{hit_sequence_ids};

  $query_fastq = $self->_detaint_filename($query_fastq);
  $bam_out = $self->_detaint_filename($bam_out);
  $ref_root = $self->_detaint_filename($ref_root);

  my $work_dir = tempdir(CLEANUP => 1);
  my $sai_out = $work_dir.q{/sai};

  #alignment
  $self->bwa_align( $ref_root, $query_fastq, $sai_out );

  #sai to sam or bam
  $args_ref ->{sai} = $sai_out;

  if($hit_sequence_ids){
     #convert sai to bam and at the same time parsing each sam line
     $self->sai2bam_parsing_se($args_ref);
  }else{
     #convert sai to sam
     $self->sai2sam_se($args_ref);
  }

  #remove sai file when sam file generated to save disk space
  unlink $sai_out;

  return 1;
}

#given two fastq and one reference, generate sam file for paired ends
sub bwa_align_pe {
  my ( $self, $args_ref ) = @_;

  my $ref_root   = $args_ref->{ref_root};
  my $fastq1     = $args_ref->{fastq1};
  my $fastq2     = $args_ref->{fastq2};
  my $sam_out    = $args_ref->{sam_out};
  my $fork_align = $args_ref->{fork_align};

  my $work_dir = tempdir(CLEANUP => 1);

  my $sai_out_1 = $work_dir.q{/1.sai};
  my $sai_out_2 = $work_dir.q{/2.sai};

  #alignments
  $self->bwa_align_list( $ref_root, [$fastq1, $fastq2], [$sai_out_1, $sai_out_2], $fork_align);

  #sai to sam
  $args_ref->{sai_1} = $sai_out_1;
  $args_ref->{sai_2} = $sai_out_2;
  $self->sai2sam_pe($args_ref);

  #remove sai files
  unlink $sai_out_1;
  unlink $sai_out_2;

  return 1;
}

#given one fastq and reference, generate sai file
sub bwa_align {
  my ( $self, $ref_root, $query_fastq, $sai_out ) = @_;

  $query_fastq = $self->_detaint_filename( $query_fastq );
  $sai_out = $self->_detaint_filename( $sai_out );
  $ref_root = $self->_detaint_filename( $ref_root );

  my $bwa_align_cmd = $self->_construct_bwa_aln_cmd( $ref_root, $query_fastq, $sai_out );

  print {*STDERR} ">Doing bwa alignment:\n $bwa_align_cmd\n" or croak $ERRNO;

  my $bwa_align_rs = system $bwa_align_cmd;
  if($bwa_align_rs != 0){
    croak "bwa align failed: $bwa_align_cmd";
  }

  return $bwa_align_cmd;
}

sub _construct_bwa_aln_cmd {
   my ($self, $ref_root, $query_fastq, $sai_out ) = @_;

   my $bwa_align_cmd = $self->bwa_cmd().q( aln );

   if( $self->bwa_options() ){
     $bwa_align_cmd .= $self->bwa_options();
   }

   $bwa_align_cmd .= qq( $ref_root $query_fastq > $sai_out);

  return $bwa_align_cmd;
}

#given a list of fastq files and one reference, generate sai files
sub bwa_align_list {
  my ( $self, $ref_root, $query_fastqs, $sai_out_files, $fork_align ) = @_;

  my $count = 0;

  if( $fork_align ){

    my $pm = Parallel::ForkManager->new( scalar @{$query_fastqs} );

    foreach my $fastq (@{$query_fastqs}){

      $count++;
      $pm->start and next;
      $self->bwa_align( $ref_root, $fastq, $sai_out_files->[$count-1] );
      $pm->finish;
    }
    $pm->wait_all_children;
  }else{

    foreach my $fastq (@{$query_fastqs}){

      $self->bwa_align( $ref_root, $fastq, $sai_out_files->[$count] );
      $count++;
    }
  }

  my $count2 = 0;
  my $bwa_aln_cmd = q{};
  foreach my $fastq (@{$query_fastqs}){
    $bwa_aln_cmd .= $self->_construct_bwa_aln_cmd( $ref_root, $fastq, $sai_out_files->[$count2]).q{;};
    $count2++;
  }

  return $bwa_aln_cmd;
}

sub sai2bam_se {
  my ( $self, $args) = @_;

  my $ref_root  = $args->{ref_root};
  my $sai       = $args->{sai};
  my $fastq     = $args->{fastq};

  my $bwa_sam = $self->bwa_cmd().qq( samse $ref_root $sai $fastq);
  my $bwa_sai2bam_cmd = $bwa_sam.q( | );
  $args->{bwa_sam} = $bwa_sam;

  my $sam_fastq_check;
  if( $self->sam_fastq_check_cmd() ){
     $sam_fastq_check = $self->sam_fastq_check_cmd() . qq{ --fastq_files $fastq };
     $args->{sam_fastq_check} = $sam_fastq_check;
     $bwa_sai2bam_cmd .= $sam_fastq_check.q{ | };
  }

  $bwa_sai2bam_cmd .= $self->_other_pipe_cmd($args);

  print {*STDERR} ">Converting single end sai file to bam:\n $bwa_sai2bam_cmd\n" or croak $ERRNO;

  my $bwa_sai2bam_rs = system "set -o pipefail; $bwa_sai2bam_cmd";

  if($bwa_sai2bam_rs != 0){
    croak "Coverting single end sai file to bam failed: $bwa_sai2bam_cmd";
  }

  return 1;
}

sub sai2bam_pe {
  my ( $self, $args ) = @_;

  my $ref_root  = $args->{ref_root};
  my $sai_1     = $args->{sai_1};
  my $sai_2     = $args->{sai_2};
  my $fastq1    = $args->{fastq1};
  my $fastq2    = $args->{fastq2};

  my $bwa_sampe = $self->bwa_cmd().qq( sampe $ref_root $sai_1 $sai_2 $fastq1 $fastq2);
  my $bwa_sai2bam_cmd = $bwa_sampe.q( | );
  $args->{bwa_sam} = $bwa_sampe;

  my $sam_fastq_check;
  if( $self->sam_fastq_check_cmd() ){
     $sam_fastq_check = $self->sam_fastq_check_cmd() . qq{ --fastq_files $fastq1 --fastq_files $fastq2};
     $args->{sam_fastq_check} = $sam_fastq_check;
     $bwa_sai2bam_cmd .= $sam_fastq_check.q{ | };
  }

  $bwa_sai2bam_cmd .= $self->_other_pipe_cmd($args);

  print {*STDERR} ">Converting paired end sai files to bam:\n $bwa_sai2bam_cmd\n" or croak $ERRNO;

  my $bwa_sai2bam_rs = system "set -o pipefail; $bwa_sai2bam_cmd";
  if($bwa_sai2bam_rs != 0){
    croak "Coverting paired end sai files to bam failed: $bwa_sai2bam_cmd";
  }

  return 1;
}

sub _other_pipe_cmd {
    my ($self, $args) = @_;

    my $id_run    = $args->{id_run};
    my $position  = $args->{position};
    my $human_split  = $args->{human_split};
    my $tag_index    = $args->{tag_index};
    my $index_fastq_file = $args->{index_fastq_file};
    my $archive_path     = $args->{archive_path};
    my $bwa_aln_cmd      = $args->{bwa_aln_cmd};
    my $bwa_sam          = $args->{bwa_sam};
    my $sam_fastq_check  = $args->{sam_fastq_check};
    my $bam_out   = $args->{bam_out};
    my $ref_root  = $args->{ref_root};

    my $other_pipe_cmd = q{};

    my $sam_original_quality;
    if($self->sam_original_quality_options()){
       $sam_original_quality = $self->sam_original_quality_cmd.q{ }.$self->sam_original_quality_options();
       $other_pipe_cmd .= $sam_original_quality .q{ | };
    }

    my $sam_second_call;
    if($self->sam_second_call_options()){
       $sam_second_call = $self->sam_second_call_cmd.q{ }.$self->sam_second_call_options();
       $other_pipe_cmd .= $sam_second_call .q{ | };
    }

    my $sam_index_tag;
    if($index_fastq_file){
       $sam_index_tag = $self->sam_index_tag_cmd.q{ --index_fastq_file }.$index_fastq_file;
       $other_pipe_cmd .= $sam_index_tag.q{ | };
    }

    my $picard_converter_cmd = $self->picard_converter_cmd();
    my $samtools_fixmate_cmd = $self->samtools_cmd().q{ fixmate - -};
    my $samtools_sort_cmd    = $self->samtools_cmd().qq{ sort - $bam_out};

    if($self->sam_header_cmd()){
       $other_pipe_cmd .= $self->_add_head_cmd({id_run      => $id_run,
                                              position      => $position,
                                              human_split   => $human_split,
                                              tag_index     => $tag_index,
                                              ref_root         => $ref_root,
                                              archive_path     => $archive_path,
                                              bwa_aln_cmd      => $bwa_aln_cmd,
                                              bwa_sam              => $bwa_sam,
                                              sam_fastq_check      => $sam_fastq_check,
                                              sam_second_call      => $sam_second_call,
                                              sam_original_quality => $sam_original_quality,
                                              sam_index_tag        => $sam_index_tag,
                                              picard_converter_cmd => $picard_converter_cmd,
                                              samtools_fixmate_cmd => $samtools_fixmate_cmd,
                                              samtool_sort_cmd     => $samtools_sort_cmd,
                                              });

    }

    $other_pipe_cmd .= qq{ | $picard_converter_cmd}
                   .qq{ | $samtools_fixmate_cmd}
                   .qq{ | $samtools_sort_cmd};

    return $other_pipe_cmd;
}


sub _add_head_cmd{
   my ($self, $args_ref) = @_;

   my $id_run = $args_ref->{id_run};
   my $position = $args_ref->{position};
   my $human_split = $args_ref->{human_split};
   my $tag_index = $args_ref->{tag_index};
   my $ref_root         = $args_ref->{ref_root};
   my $archive_path     = $args_ref->{archive_path};
   my $bwa_aln_cmd      = $args_ref->{bwa_aln_cmd};
   my $bwa_sam          = $args_ref->{bwa_sam},
   my $sam_fastq_check  = $args_ref->{sam_fastq_check},
   my $sam_original_quality  = $args_ref->{sam_original_quality},
   my $sam_second_call       = $args_ref->{sam_second_call},
   my $sam_index_tag         = $args_ref->{sam_index_tag};
   my $picard_converter_cmd  = $args_ref->{picard_converter_cmd};

   # this two commands not pass to header, construct them again using smatools command
   my $samtools_fixmate_cmd  = $args_ref->{samtools_fixmate_cmd};
   my $samtools_sort_cmd     = $args_ref->{samtools_sort_cmd};

   my $add_header_cmd = $self->sam_header_cmd().q{ --id_run }.$id_run;

   if($position){
      $add_header_cmd .= qq{ --position $position};
   }

   if( $archive_path ){
      $add_header_cmd .= q{ --archive_path }.$archive_path;
   }

   if($human_split){
      $add_header_cmd .= qq{ --human_split $human_split};
   }

   if(defined $tag_index){
      $add_header_cmd .= qq{ --tag_index $tag_index};
   }

   $add_header_cmd .= q{ --bwa_cmd }.$self->bwa_cmd();
   $add_header_cmd .= q{ --bwa_aln_cmd '}.$bwa_aln_cmd.q{'};
   $add_header_cmd .= q{ --bwa_sam_cmd '}.$bwa_sam.q{'};

   if( $sam_fastq_check ){
     $add_header_cmd .= q{ --sam_fastq_check_cmd '}.$sam_fastq_check.q{'};
   }

   if($self->sam_original_quality_options()){
     $add_header_cmd .= q{ --sam_original_quality_cmd '}.$sam_original_quality.q{'};
   }

   if($self->sam_second_call_options()){
     $add_header_cmd .= q{ --sam_second_call_cmd '}.$sam_second_call.q{'};
   }

   if( $sam_index_tag ){
      $add_header_cmd .= q{ --sam_index_tag_cmd '}.$sam_index_tag.q{'};
   }

   $add_header_cmd .= q{ --picard_converter_cmd '}.$picard_converter_cmd.q{'};

   $add_header_cmd .= q{ --samtools_cmd }.$self->samtools_cmd();

   return $add_header_cmd;
}

sub sai2sam_se {
  my ( $self, $args ) = @_;

  my $ref_root  = $args->{ref_root};
  my $sai       = $args->{sai};
  my $fastq     = $args->{fastq};
  my $sam_out   = $args->{bam_out};

  my $bwa_sam_cmd = $self->bwa_cmd().qq( samse $ref_root $sai $fastq > $sam_out );

  print {*STDERR} ">bwa converting single end sai files to sam:\n $bwa_sam_cmd\n" or croak $ERRNO;

  my $bwa_sam_rs = system $bwa_sam_cmd;
  if($bwa_sam_rs != 0){
    croak "bwa align failed: $bwa_sam_cmd";
  }

  return 1;
}

sub sai2bam_parsing_se {
  my ( $self, $args ) = @_;

  my $ref_root  = $args->{ref_root};
  my $sai       = $args->{sai};
  my $fastq     = $args->{fastq};
  my $bam_out   = $args->{bam_out};
  my $hit_sequence_ids     = $args->{hit_sequence_ids};
  my $nonhit_sequence_ids = $args->{nonhit_sequence_ids};

  my $bwa_sam_cmd        = $self->bwa_cmd().qq( samse $ref_root $sai $fastq |);
  my $samtools_samse_cmd = q{| }.$self->samtools_cmd().qq( view -b -S -o $bam_out - );

  print {*STDERR} ">bwa converting single end sai files to sam:\n $bwa_sam_cmd\n" or croak $ERRNO;
  open my $sam_out_fh, $bwa_sam_cmd or croak $ERRNO; ## no critic (InputOutput::RequireBriefOpen InputOutput::ProhibitTwoArgOpen)

  print {*STDERR} ">samtools converting sam file to bam:\n $samtools_samse_cmd\n" or croak $ERRNO;
  open my $sam_in_fh, $samtools_samse_cmd or croak $ERRNO; ## no critic (InputOutput::RequireBriefOpen InputOutput::ProhibitTwoArgOpen)

  while(my $line = <$sam_out_fh>){
       print {$sam_in_fh} $line or croak $ERRNO;

       chomp $line;

       #ignore header section
       if($line =~ /^@/mxs){
         next;
       }
       my @fields_line = split /\t/sxm, $line;

       my $sequence_id = $fields_line[0];
       my $rname =  $fields_line[2];

       if($rname eq q{*}){

            $nonhit_sequence_ids->{$sequence_id}++;
       }else{

           $hit_sequence_ids->{$sequence_id}++;
       }
  }

  close $sam_out_fh or croak $ERRNO;
  close $sam_in_fh or croak $ERRNO;

  return 1;
}

sub sai2sam_pe {
  my ( $self, $args ) = @_;

  my $ref_root  = $args->{ref_root};
  my $sai_1     = $args->{sai_1};
  my $sai_2     = $args->{sai_2};
  my $fastq1    = $args->{fastq1};
  my $fastq2    = $args->{fastq2};
  my $sam_out   = $args->{sam_out};

  my $bwa_sam_cmd = $self->bwa_cmd().qq( sampe $ref_root $sai_1 $sai_2 $fastq1 $fastq2 > $sam_out );

  print {*STDERR} ">bwa converting paired end sai files to sam:\n $bwa_sam_cmd\n" or croak $ERRNO;

  my $bwa_sam_rs = system $bwa_sam_cmd;
  if($bwa_sam_rs != 0){
    croak "bwa align failed: $bwa_sam_cmd";
  }

  return 1;
}

sub sam2bam {
  my ($self, $sam_in, $bam_out, $ref_fai) = @_;

  my $sam2bam_cmd = $self->samtools_cmd().qq( import $ref_fai $sam_in $bam_out);

  print {*STDERR} ">Converting sam to bam\n $sam2bam_cmd\n" or croak $ERRNO;

  my $sam2bam_cmd_rs = system $sam2bam_cmd;

  if($sam2bam_cmd_rs != 0){

    croak 'sam2bam failed';
  }

  return 1;
}

sub sort_bam {
  my ($self, $bam_in, $bam_out) = @_;

  my $sort_bam_cmd = $self->samtools_cmd().qq( sort $bam_in $bam_out);

  print {*STDERR} ">Sorting bam\n $sort_bam_cmd\n" or croak $ERRNO;

  if(! -e $bam_in){
    croak "The bam file to be sorted $bam_in not exist";
  }

  my $sort_bam_cmd_rs = system $sort_bam_cmd;

  if($sort_bam_cmd_rs != 0){

    croak 'sort bam failed';
  }

  return 1;
}

sub parsing_pileup_bam{
  my ($self, $sorted_bam, $ref_fa, $pileup_filename) = @_;

  my $pileup_cmd = $self->samtools_cmd().qq( mpileup -f $ref_fa $sorted_bam);
  my $chromosome_coverage = {};
  my $chromosome_depth = {};

  print {*STDERR} ">Start doing pileup from sorted bam\n" or croak $ERRNO;
  print {*STDERR} "$pileup_cmd\n" or croak $ERRNO;

  my $pileup_err = Symbol::gensym;
  my $pid = open3( undef, my $pileup_fh, $pileup_err, $pileup_cmd );

  #open my $pileup_output, q{>}, qq{$pileup_filename} or croak $ERRNO;## no critic (InputOutput::RequireBriefOpen)

  while(my $line = <$pileup_fh>){

    chomp $line;
    my @fields = split /\s+/msx, $line;
    my $chromosome = shift @fields;
    my $depth = $fields[2];
    #print {$pileup_output} $line, "\n" or croak $ERRNO;
    $chromosome_coverage->{$chromosome}++;
    $chromosome_depth->{$chromosome}->{$depth}++;
    $chromosome_depth->{q{All}}->{$depth}++;
  }
  #close $pileup_output or croak $ERRNO;
  close $pileup_fh or croak $ERRNO;
  close $pileup_err or croak $ERRNO;

  waitpid $pid, 0;
  return ($chromosome_coverage, $chromosome_depth);
}

sub alignment_coverage_from_sam{
  my ($self, $sorted_bam, $ref_fa) = @_;

  my $ref_fai = $ref_fa.q{.fai};
  my $work_dir = tempdir(CLEANUP => 1);

  my $pileup_filename = $work_dir.q{/pileup.txt};

  my ($chromosome_coverage, $chromosome_depth) = $self->parsing_pileup_bam($sorted_bam, $ref_fa, $pileup_filename);

  unlink $sorted_bam;

  my $reference_length = $self->count_reference_length($ref_fa);

  my $total_coverage;
  my $total_reference_length;

  foreach my $chromosome (keys %{$chromosome_coverage}){

    $total_coverage += $chromosome_coverage->{$chromosome};
    $total_reference_length += $reference_length->{$chromosome};

    if($reference_length->{$chromosome}){

      $chromosome_coverage->{$chromosome}
      =  sprintf '%.4f', $chromosome_coverage->{$chromosome}/$reference_length->{$chromosome};
    }else{

      $chromosome_coverage->{$chromosome} = 0;
    }

  }

  if($total_reference_length){

    $total_coverage = sprintf '%.4f', $total_coverage/$total_reference_length;
  }else{

    $total_coverage = 0;
  }
  $chromosome_coverage->{q{All}} = $total_coverage;

  return ($chromosome_coverage, $chromosome_depth);
}

sub count_reference_length{
  my ($self, $ref_fa) = @_;

  open my $ref_fa_fh, '<', $ref_fa or croak $ERRNO; ## no critic (InputOutput::RequireBriefOpen)

  my %reference_length = ();
  my $length;
  my $chromosome;

  while (my $line = <$ref_fa_fh>){

    chomp $line;
    if($line =~ /^>(\S+)\s+/msx){
      if(defined $chromosome){

        $reference_length{$chromosome} = $length;
      }

      $chromosome = $1;
      $length = 0;
    }else{
      $length += length $line;
    }
  }

  close $ref_fa_fh or croak $ERRNO;

  if(defined $chromosome){
      $reference_length{$chromosome} = $length;
  }

  return \%reference_length;
}

sub _detaint_filename {
  my ($self, $string) = @_;

  my ($dt) = $string =~ m{^ ( [[:word:]./\\+#-]+ ) $}msx;
  if ($dt ne $string) {

    croak "could not detaint $string, got $dt";
  }

  return $dt;
}
1;

__END__

=head1 NAME

npg_common::Alignment

=head1 VERSION

$LastChangedRevision: 12834 $

=head1 SYNOPSIS

  my $oBWAAlignment = npg_common::Alignment->new({bwa_cmd => q{bwa},
                                                   samtools_cmd => q{samtools},
                                                 );
  $oBWAAlignment->bwa_align_se($query_fastq, $sam, $reference_bwa);

  $oBWAAlignment->bwa_align_pe({fastq1=>$query_fastq1, fastq2=>$query_fastq2, sam_out=>$sam, fork_align=>1, ref_root=>$reference_bwa});

=head1 DESCRIPTION

Class to do bwa alignment based on the fastq file

=head1 SUBROUTINES/METHODS

=head2 new - constructor to create a object

=head2 sam2bam - covert sam to bam format

=head2 sort_bam - sort the bam file

=head2 parsing_pileup_bam - given a bam file, generate pileup format and count the alignment coverage and depth

=head2 alignment_coverage_from_sam - given a sam file, generate pileup format and calculate the alignment coverage and depth

=head2 count_reference_length - check the reference length

=head2 bwa_align - given one fastq file and reference location, doing bwa alignment and generate sai file

=head2 bwa_align_list - given a list of fastq files and output sai file list, doing bwa alignment

=head2 bwa_align_se - given one fastq file and reference location, doing bwa alignment and output in sam format

=head2 bwa_align_pe - given two fastq files and reference location, doing bwa alignment and generate one sam file. The two alignments could be done in parallel.

=head2 sai2sam_se - bwa convert sai file to sam file

=head2 sai2bam_parsing_se - bwa convert sai file to bam file and parsing each sam line to store hit and nonhit reads id into a hash

=head2 sai2sam_pe - given two paired read sai file, convert them to sam file

=head2 bwa_align_bam_se - doing bwa alignment and output bam

=head2 bwa_align_bam_pe - doing bwa alignment and output bam

=head2 sai2bam_se - convert sai file to bam

=head2 sai2bam_pe - convert sai file to bam

=head2 _detaint_filename

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Carp

=item English

=item Class::Accessor

=item File::Temp qw(tempfile tempdir)

=item IPC::Open3

=item Parallel::ForkManager

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Guoying Qi, E<lt>gq1@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2009 GRL, by Guoying Qi

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.8 or,
at your option, any later version of Perl 5 you may have available.

=cut
