#########
# Author:        gq1
# Maintainer:    $Author: gq1 $
# Created:       2011-08-02
# Last Modified: $Date: 2012-02-23 10:06:50 +0000 (Thu, 23 Feb 2012) $
# Id:            $Id: bam_align_irods.pm 15229 2012-02-23 10:06:50Z gq1 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-43.0/lib/npg_common/bam_align_irods.pm $

package npg_common::bam_align_irods;

use Moose;
with qw(
    MooseX::Getopt npg_common::roles::log
);

use Carp;
use English qw(-no_match_vars);
use File::Temp qw( tempfile tempdir );

use npg_common::irods::Loader;
use npg_common::irods::run::Bam;
use npg_common::sequence::reference;
use npg_qc::Schema;
use File::Basename;
use Perl6::Slurp;
use IPC::Open3;
use JSON;
use File::Copy;

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 15229 $ =~ /(\d+)/msx; $r; };

Readonly::Scalar my $IRODS_PATH      => '/software/irods/icommands/bin/';
Readonly::Scalar my $BAM_ALIGN_CMD   => '/software/solexa/src/new-pipeline-dev/useful_modules/trunk/scripts/bam_aligner';
Readonly::Scalar my $BAM_FLAGSTATS_LOADER_CMD  => '/software/solexa/bin/npg_qc_autoqc_data.pl';
Readonly::Scalar my $JAVA_CMD                  => '/software/bin/java';
Readonly::Scalar my $CHANGE_BAM_HEADER_JAR     => '/software/solexa/bin/aligners/illumina2bam/current/ChangeBamHeader.jar';
Readonly::Scalar my $BAM_ALIGNMENT_CMD         => '/software/solexa/bin/bam_alignment.pl';
Readonly::Scalar my $BAM_MARKDUPLICATES_CMD    => '/software/solexa/bin/bam_mark_duplicate.pl';
Readonly::Scalar my $PICARD_SORT_SAM_JAR       => '/software/solexa/bin/aligners/picard/current/SortSam.jar';
Readonly::Scalar my $STRIP_BAM_TAG_JAR         => '/software/solexa/bin/aligners/illumina2bam/current/BamTagStripper.jar';

Readonly::Scalar our $SAMTOOLS => q{/software/solexa/bin/aligners/samtools/current/samtools};

Readonly::Scalar my $DEFAULT_TAG_INDEX => -1;
Readonly::Scalar our $EXIT_CODE_SHIFT => 8;

sub run {
   my $self = shift;

   my $task = $self->task();

   $self->get_old_bam_from_irods();

   if( $task eq q{non_consented_split} ){
      $self->sort_old_bam();
   }

   $self->$task();

   $self->reload_bam();
   if( $task eq q{non_consented_split} ){

      $self->reload_bam('human');
   }

   if( $task ne q{change_bam_header} &&  $task ne q{strip_bam_tag}){

      $self->delete_qc_bam_flagstats();

      $self->reload_bam_flagstats();
   }

   return;
}

has 'id_run' => (
    is            => 'ro',
    isa           => 'Int',
    required      => 1,
);

has 'lane' => (
    is            => 'ro',
    isa           => 'Int',
    required      => 1,
);

has 'tag_index' => (
    is            => 'ro',
    isa           => 'Int',
    required      => 0,
);

has 'alignment_filter' => (isa        => 'Str',
                           is         => 'rw',
                           required   => 0,
                          );

has 'rt_ticket' => (
    is            => 'ro',
    isa           => 'Int',
    required      => 0,
);

has 'task' => (
    is            => 'ro',
    isa           => 'Str',
    required      => 1,
);

has 'reference' => (
    is            => 'ro',
    isa           => 'Str',
    lazy_build    => 1,
);

sub _build_reference {
   my $self = shift;
   my $ref_finder = npg_common::sequence::reference->new(id_run=> $self->id_run(),
                                                         position => $self->lane(),
                                                         tag_index => $self->tag_index(),
   );
   my $refs = $ref_finder->refs;
   if(scalar @{$refs} >1){
     croak 'more than one reference found';
   }
   return $refs->[0];
}

has 'human_reference'   => (isa           => 'Str',
                            is            => 'rw',
                            required      => 0,
                            lazy_build    => 1,
                            documentation => 'A path to human reference for bwa alignment to split non-consented reads',
                           );

sub _build_human_reference {
   my $self = shift;

   my $ruser = Moose::Meta::Class->create_anon_class(
          roles => [qw/npg_common::sequence::reference::roles::find/])
          ->new_object({species => q{Homo_sapiens}});

   my $human_ref;
   eval {
      $human_ref = $ruser->refs->[0];
      1;
   } or do{
      carp $EVAL_ERROR;
   };

   return $human_ref;
}

has  '_lims'    =>  ( isa             => 'Maybe[st::api::lims]',
                      is              => 'ro',
                      init_arg        => undef,
                      lazy_build      => 1,
                    );
sub _build__lims {
    my $self = shift;

    if( defined $self->id_run() ){
       return st::api::lims->new(id_run  => $self->id_run(),
                            position  => $self->lane(),
                            tag_index => $self->tag_index(),
                            );
    }

    $self->log( 'id_run, position or tag_index need to be set to get lims information' );
    return;
}


has  '_npg_api_run'    =>  ( isa           => 'Maybe[npg::api::run]',
                             is            => 'ro',
                             init_arg        => undef,
                             lazy_build      => 1,
                           );
sub _build__npg_api_run {
    my $self = shift;

    my $st_lims = $self->_lims();
    if( ! defined $st_lims ){
       return $st_lims;
    }
    return $st_lims->npg_api_run();
}

has 'working_dir' => (
    is            => 'ro',
    isa           => 'Str',
    required      => 1,
);

has 'bam_filename' => (
    is            => 'ro',
    isa           => 'Str',
    lazy_build    => 1,
);
sub _build_bam_filename {
   my $self = shift;

   my $name = $self->id_run().q{_}.$self->lane();

   if(defined $self->tag_index() ){
      $name .= q{#}.$self->tag_index();
   }
   if( $self->alignment_filter() ){
      $name .= q{_}.$self->alignment_filter();
   }
   return $name.q{.bam};
}

has 'old_bam_filename' => (
    is            => 'ro',
    isa           => 'Str',
    lazy_build    => 1,
);
sub _build_old_bam_filename {
   my $self = shift;

   my $bam_filename = $self->bam_filename();

   $bam_filename =~ s/[.]bam$/_old\.bam/mxs;

   return $bam_filename;
}

sub get_old_bam_from_irods {
   my $self = shift;

   mkdir $self->working_dir();
   my $command = $IRODS_PATH.q{iget -f -K /seq/}.$self->id_run().q{/}.$self->bam_filename(). q{ }. $self->working_dir().q{/}.$self->old_bam_filename();
   $self->log($command);
   my $iget_rs =  system $command;
   if($iget_rs){
      croak "iget failed: $command";
   }
   return;
}

sub sort_old_bam {
   my $self = shift;

   my $old_bam =  $self->working_dir().q{/}.$self->old_bam_filename();
   my $unsorted_old_bam = $self->working_dir().q{/}.$self->old_bam_filename();
   $unsorted_old_bam =~ s/[.]bam$/_unsorted.bam/mxs;

   $self->log("move  $old_bam, $unsorted_old_bam");
   move  $old_bam, $unsorted_old_bam;

   my $sort_cmd = qq{$JAVA_CMD -Xmx3000m -jar $PICARD_SORT_SAM_JAR}
               .qq{ I=$unsorted_old_bam}
               .qq{ O=$old_bam}
               .q{ VALIDATION_STRINGENCY=SILENT SORT_ORDER=queryname};
   $self->log($sort_cmd);
   my $sort_cmd_rs = system $sort_cmd;
   if($sort_cmd_rs){
      croak "sort sam failed: $sort_cmd";
   }
   return;
}

sub bam_realign {
   my $self = shift;

   my $old_bam = $self->working_dir().q{/}.$self->old_bam_filename();
   my $new_bam = $self->working_dir().q{/}.$self->bam_filename();
   my $scratch =  tempdir( DIR => $self->working_dir(), CLEANUP => 1 );

   my $command = q{cat }. $old_bam
                 . q{ | }
                 . $BAM_ALIGN_CMD
                 . q{ --scratch }.$scratch
                 . q{ --index_base }.$self->reference()
                 . q{ --out_bam }.$new_bam;

   if( $self->id_run() ){
        $command .= q{ --id_run } . $self->id_run();
   }
   if( $self->lane() ){
        $command .= q{ --position } . $self->lane();
   }
   if( defined $self->tag_index() ){
        $command .= q{ --tag_index } . $self->tag_index();
   }
   if( $self->alignment_filter() ){
        $command .= q{ --alignment_filter } . $self->alignment_filter();
   }

   $self->log($command);

   my $iget_rs = system "set -o pipefail; $command";
   if($iget_rs){
      croak "realignment failed: $command";
   }
   return;
}

sub reload_bam {
    my ( $self, $human_split ) = @_;

    my $bam_filename = $self->bam_filename();
    if( $human_split ) {
        $bam_filename =~ s/[.]bam$/_human.bam/mxs;
    }

    my $file = $self->working_dir().q{/}.$bam_filename;

    my $meta = {};
    if( $self->task() eq q{bam_realign} ){
       $meta = {
                reference  => $self->reference(),
                alignment  => 1,
               };
    }elsif ( $self->task() eq q{non_consented_split} ){
       my $reference = $self->reference();
       $meta->{target} = 1;
       if( $human_split ) {
          $reference = $self->human_reference();
          $meta->{target} = 0;
          $meta->{id_run} = $self->id_run();
          $meta->{lane}   = $self->lane();
          $meta->{type}   = q{bam};
          $meta->{alignment_filter} = q{human};
       }
       $meta->{reference} = $reference;
       if( $reference ){
          $meta->{alignment} = 1;
       }
       my $total_reads = $self->_get_number_of_reads_qc($file);
       if( defined $total_reads ) {
           $meta->{total_reads} = $total_reads;
       }
    }

    if($self->rt_ticket()){
        $meta->{rt_ticket} = $self->rt_ticket();
    }

    my $loader = npg_common::irods::Loader->new({
       file         => $file,
       meta_data    => $meta,
       new_filename => $bam_filename,
       collection   => '/seq/'.$self->id_run(),
    });
    $loader->run();
    my $irods_file = $loader->collection().q{/}.$loader->new_filename();
    $self->remove_obsolete_replication($irods_file);

    my $bam_header = npg_common::irods::run::Bam->new(id_run => $self->id_run())->get_bam_reference($file);
    my $alignment = $bam_header->{alignment};
    if(!$alignment){
       $self->log('Bam file not aligned, no need to load bai file');
       return 1;
    }

    $meta = {type => 'bai'};
    $file =~ s/bam$/bai/mxs;
    my $irods_bai = $bam_filename;
    $irods_bai =~ s/bam$/bai/mxs;
    $loader = npg_common::irods::Loader->new({
       file         => $file,
       meta_data    => $meta,
       new_filename => $irods_bai,
       collection   => '/seq/'.$self->id_run(),
    });
    $loader->run();
    $irods_file = $loader->collection().q{/}.$loader->new_filename();
    $self->remove_obsolete_replication($irods_file);
    return;
}

sub remove_obsolete_replication {
   my ($self, $bam_file) = @_;

   my $rep_num_to_remove = $self->_get_rep_num_to_remove($bam_file);
   foreach my $rep_num (@{$rep_num_to_remove}){

   	 my $irm_command = $IRODS_PATH . qq{irm -n $rep_num $bam_file};
   	 $self->log($irm_command);
   	 my $irm_rs =  system $irm_command;
       if($irm_rs){
          croak "Failed: $irm_command";
       }
   }
   return;
}

sub _get_rep_num_to_remove {
    my ($self, $bam_file) = @_;

    my @rep_num_to_remove = ();

    my $ils_cmd = $IRODS_PATH . qq{ils -l $bam_file};

    my $pid = open3( undef, my $ils_out_fh, undef, $ils_cmd);

    while (my $line = <$ils_out_fh> ) {

       chomp $line;

       if ( $line =~ /ERROR[:][ ]lsUtil[:][ ]/mxs ){
          #file not exists in the server
          last;
       }elsif ( $line !~ /[ ]&[ ]/mxs ){
          my ($rep_num) = $line =~ /^[ ]{2}\w+\s+(\d+)/mxs;
          if(defined $rep_num){
             push @rep_num_to_remove, $rep_num;
          }
       }
    }

    waitpid $pid, 0;

    if( $CHILD_ERROR >> $EXIT_CODE_SHIFT ){
       carp "Problems to list bam file $bam_file";
     }

    close $ils_out_fh or croak "can not close ils command output: $ERRNO";

    return \@rep_num_to_remove;
}


sub delete_qc_bam_flagstats {
   my $self = shift;

   my $schema = npg_qc::Schema->connect( );

   my $query = {id_run => $self->id_run(), position => $self->lane()};
   if( defined $self->tag_index() ){
       $query->{tag_index} = $self->tag_index();
   }else {
       $query->{tag_index} = $DEFAULT_TAG_INDEX;
   }
   if( $self->alignment_filter() ){
      $query->{human_split} = $self->alignment_filter();
   }else {
      $query->{human_split} = q{all};
   }
   $self->log('Deleting npgqc bam flagstats for run '
             . $query->{id_run}
             . q{ lane } . $query->{position}
             . q{ tag } . $query->{tag_index}
             . q{ human_split } . $query->{human_split}
           );
   $schema->txn_do(
       sub {
           my $result = $schema->resultset('BamFlagstat')
                         ->search( $query )->delete_all();
           if($result){
              $self->log('deletion successfull');
           }
       }
   );

   return;
}

sub reload_bam_flagstats {
   my $self = shift;

   my $command = $BAM_FLAGSTATS_LOADER_CMD . q{ --id_run } . $self->id_run()
                 . q{ --path } . $self->working_dir();
   $self->log($command);
   my $loader_rs =  system $command;
   if($loader_rs){
      croak "reload bam flagstats failed: $command";
   }
   return;
}

sub change_bam_header {
    my $self = shift;

    $self->log('geting sample, library, and study informtation');

    my $lims = $self->_lims();
    my $sample_name = $lims->sample_name();
    my $library_id = $lims->library_id();
    my $study_name = $lims->study_name();
    my $study_description = $lims->study_description();

    my $old_bam = $self->working_dir().q{/}.$self->old_bam_filename();
    my $new_bam = $self->working_dir().q{/}.$self->bam_filename();

    my $change_header_cmd = qq{$JAVA_CMD -Xmx1024m -jar $CHANGE_BAM_HEADER_JAR VALIDATION_STRINGENCY=SILENT CREATE_INDEX=true CREATE_MD5_FILE=true I=$old_bam}
                          . qq{ O=$new_bam};
    if($sample_name){
        $change_header_cmd .= q{ SM="} . $self->_check_tag_value($sample_name) . q{"};
    }
    if($library_id){
       $change_header_cmd .= q{ LB="} . $self->_check_tag_value($library_id) . q{"};
    }
    if($study_name){
        $change_header_cmd .= q{ DS="} . $self->_check_tag_value($study_name);
        if($study_description){
           $change_header_cmd .= q{: }. $self->_check_tag_value($study_description);
        }
        $change_header_cmd .= q{"};
    }

    $self->log($change_header_cmd);
    my $change_rs =  system $change_header_cmd;
    if($change_rs){
      croak "change bam header failed: $change_header_cmd";
    }
}


sub strip_bam_tag {
    my $self = shift;

    my $old_bam = $self->working_dir().q{/}.$self->old_bam_filename();
    my $new_bam = $self->working_dir().q{/}.$self->bam_filename();

    my $strip_bam_tag_cmd = qq{$JAVA_CMD -Xmx1024m -jar $STRIP_BAM_TAG_JAR VALIDATION_STRINGENCY=SILENT CREATE_INDEX=true CREATE_MD5_FILE=true I=$old_bam}
                          . qq{ O=$new_bam}
                          . q{ STRIP=OQ};

    $self->log($strip_bam_tag_cmd);
    my $strip_rs =  system $strip_bam_tag_cmd;
    if($strip_rs){
      croak "Strip bam tag failed: $strip_bam_tag_cmd";
    }
}

sub _check_tag_value {
  my ($self, $tag_value) = @_;

  $tag_value =~ s/\n/\ /gmxs;
  $tag_value =~ s/\t/\ /gmxs;

  return $tag_value;
}


sub non_consented_split {
   my $self = shift;

   my $is_paired_read = $self->_npg_api_run->is_paired_read();
   my $old_bam = $self->working_dir().q{/}.$self->old_bam_filename();
   my $output_prefix =  $self->working_dir().q{/}. $self->id_run().q{_}.$self->lane();
   if(defined $self->tag_index() ){
      $output_prefix .= q{#}.$self->tag_index();
   }
   my $reference = $self->reference();
   my $ref_dict = $reference;
   $ref_dict =~ s/bwa/picard/mxs;
   $ref_dict .= q{.dict};

   my $bam_align_filter_cmd = $BAM_ALIGNMENT_CMD . qq{ --input $old_bam --output_prefix $output_prefix --non_consent_split --reference $reference --ref_dict $ref_dict};
   if( $is_paired_read ) {
       $bam_align_filter_cmd  .= q{ --is_paired_read};
   }

   $self->log( $bam_align_filter_cmd );
   my $bam_alignment_rs = system $bam_align_filter_cmd;
   if( $bam_alignment_rs ) {
      croak "bam alignment and filtering failed: $bam_align_filter_cmd";
   }

   $self->_mark_duplicates($output_prefix);

   $self->_mark_duplicates($output_prefix, q{human});

   return;
}

sub _mark_duplicates {
   my ($self, $bam_prefix, $human_split) = @_;

   if( $human_split ){
       $bam_prefix .= q{_human};
   }

   my $input_bam  = $bam_prefix . q{.bam};
   my $output_bam = $bam_prefix . q{_mk.bam};
   my $metrics    = $bam_prefix . q{_bam_flagstats.json};

   my $mk_cmd = $BAM_MARKDUPLICATES_CMD
              . qq{ --input_bam $input_bam --output_bam $output_bam --metrics_json $metrics}

              . q{ --id_run } .$self->id_run() . q{ --position } . $self->lane();
   if( defined $self->tag_index() ){
       $mk_cmd .= q{ --tag_index } . $self->tag_index();
   }
   if( $human_split ){
       $mk_cmd .= q{ --human_split human};
   }

   $mk_cmd .= q{ --sort_input --change_bam_header  --replace_file};

   $self->log( $mk_cmd );
   my $mk_duplicates_rs =  system $mk_cmd;
   if( $mk_duplicates_rs ){
      croak "mark duplicates failed: $mk_cmd";
   }

   return;

}

sub _get_number_of_reads_qc {
   my ($self, $bam) = @_;

   my $num_total_reads;

   my ($name, $path, $suffix) = fileparse($bam, q{.bam});

   my $bam_flagstats_file = File::Spec->catfile ( $path, $name.q{_bam_flagstats.json} );

   if(-e $bam_flagstats_file ){

      my $flag_stats = from_json(slurp $bam_flagstats_file);
      $num_total_reads = $flag_stats->{num_total_reads};
   }

   return $num_total_reads;
}

no Moose;

__PACKAGE__->meta->make_immutable;

1;

__END__

=head1 NAME

npg_common::bam_align_irods

=head1 VERSION

$Revision: 15229 $

=head1 SYNOPSIS

use npg_common::bam_align_irods;

my $bam_re_aln = npg_common::bam_align_irods->new_with_options();

$bam_re_aln->run();

=head1 DESCRIPTION

This moudule will download the irods file to the working directory, do the re-alignment and reload them back to irods

=head1 SUBROUTINES/METHODS

=head2 run

Main method. Control calling of other methods.

=head2 get_old_bam_from_irods

=head2 sort_old_bam

=head2 bam_realign

=head2 reload_bam

=head2 remove_obsolete_replication

=head2 change_bam_header

=head2 strip_bam_tag

=head2 non_consented_split

reload new aligned bam file to irods with bai file as well

=head2 delete_qc_bam_flagstats

=head2 reload_bam_flagstats

=head1 DIAGNOSTICS

=head1 DEPENDENCIES

=head1 CONFIGURATION AND ENVIRONMENT

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Guoying Qi, E<lt>gq1@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2011 GRL, by Guoying Qi

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
