#########
# Author:        jo3
# Maintainer:    $Author: dj3 $
# Created:
# Last Modified: $Date: 2012-02-29 13:11:34 +0000 (Wed, 29 Feb 2012) $
# Id:            $Id: aligners.pm 15261 2012-02-29 13:11:34Z dj3 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-43.0/lib/npg_common/roles/aligners.pm $

package npg_common::roles::aligners;

use Moose::Role;

use Carp;
use File::Spec::Functions;
use IPC::Open3;
use Perl6::Slurp;
use autodie qw(:all);
use Cwd qw(cwd abs_path);

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 15261 $ =~ /(\d+)/msx; $r; };

Readonly::Scalar my $NPD_COLLECTION => '/software/solexa/bin/aligners';

# Update this with aligners that require a custom regex to extract the version
# string from their output.
Readonly::Hash my %VERSION_REGEX => (
    stampy => qr{^ stampy \s+ (\S+ (?: \s [(]\S+[)] )? ) }imsx,
    abyss  => qr{^ ABYSS \s+ [(]ABySS[)] \s+ (\S+) }imsx,
    ssaha2 => qr{^= \s+ SSAHA2 \s+ version \s (\S+) }imsx,
);

# Update this with aligners that do not have an executable in
# $self->collection_root/bin with the same name as the aligner package.
Readonly::Hash my %ALIGN_COMMAND => (
    snpomatic => 'findknownsnps',
    abyss     => 'ABYSS',
    soap2     => 'soap',
    fuzzypath => 'fuzzypath.csh',
);


has 'collection_root' => (
    is            => 'ro',
    isa           => 'Str',
    default       => $NPD_COLLECTION,
    documentation => 'Where aligners are installed',
);


has 'collection_bin' => (
    is            => 'ro',
    isa           => 'Str',
    default       => catfile( $NPD_COLLECTION, 'bin' ),
    documentation => 'Symlinks to latest version of executables in here',
);


has '_aligner_list' => (
    is            => 'ro',
    isa           => 'ArrayRef[Str]',
    reader        => 'aligner_list',
    lazy_build    => 1,
    documentation =>
        'The list of installed aligners/assemblers/alignment tools',
);

sub _build__aligner_list {
    my ($self) = @_;

    my $root = $self->collection_root();

    my @list;

    my $dir_handle;
    opendir $dir_handle, $root;
    foreach my $entry ( readdir $dir_handle ) {
        next if $entry =~ m/^ [.] /msx;
        next if $entry eq 'bin';
        next if !-d "$root/$entry";

        push @list, $entry;
    }
    closedir $dir_handle;

    return \@list;
}


sub current_base_dir {
    my ( $self, $aligner_name ) = @_;
    croak 'Aligner name required as argument' if !$aligner_name;

    my $dir;

    if ( $aligner_name eq 'abyss' ) {
        $dir = catdir( $self->collection_root(), $aligner_name,
                        'current', 'installed', 'bin' );
    }
    else {
        $dir = catdir( $self->collection_root(), $aligner_name, 'current' );
    }

    if ( !-d $dir ) {
        carp "$dir not found";
    }



    return abs_path($dir);
}


sub current_version {
    my ( $self, $aligner_name ) = @_;
    croak 'Aligner name required as argument' if !$aligner_name;


    my $command = ( defined $ALIGN_COMMAND{$aligner_name} )
                ? $ALIGN_COMMAND{$aligner_name}
                : $aligner_name;

    my $root        = $self->collection_root();
    my $aligner_cmd = catfile( $root, 'bin', $command );


    my $default_regex = qr{^
                            (?: $aligner_cmd )? \s*
                            version [:]? \s+
                            (\S+ (?: [ \t]+ \S+ )? )
                          }imsx;

    my $version_regex = ( defined $VERSION_REGEX{$aligner_name} )
                      ? $VERSION_REGEX{$aligner_name}
                      : $default_regex;


    # Special pleading for some aligners/tools.
    my $temp_path = $ENV{PATH};
    if ( $aligner_name eq q{stampy} ) {
        $aligner_cmd .= q{.py};

        # Special special pleading for WTSI.
        $temp_path = '/software/python-2.7/bin:' . $ENV{PATH};
    }

    ( $aligner_name eq q{velvet} ) && ( $aligner_cmd .= q{h} );
    ( $aligner_name eq q{picard} ) && ( $aligner_cmd = q{/bin/true} );

    croak "$aligner_cmd not found" if !-e $aligner_cmd;

    local $ENV{PATH} = $temp_path;
    my $aligner_version;
    foreach my $arg ( q{}, '--version', '-v', '-V', 'version' ) {

        my $out;
        my $pid = open3( undef, $out, $out, "$aligner_cmd $arg" );
        waitpid $pid, 0;

        my $output = slurp($out);

        ($aligner_version) = $output =~ m/$version_regex/igmsx;
        last if defined $aligner_version;
    }

    if ( !defined $aligner_version ) {
        my $current_symlink = catfile( $root, $aligner_name, 'current' );

        $aligner_version = ( -e $current_symlink )
                         ? readlink($current_symlink)
                         : q{No version string obtained};
    }

    return $aligner_version;
}


sub installed_versions {
    my ( $self, $aligner_name ) = @_;
    croak 'Aligner name required as argument' if !$aligner_name;

    my $root        = $self->collection_root();
    my $aligner_dir = catdir( $root, $aligner_name );
    croak "$aligner_name not maintained in $root" if !-d $aligner_dir;

    my @list;

    my $dir_handle;
    opendir $dir_handle, $aligner_dir;
    foreach my $entry ( readdir $dir_handle ) {
        my $path = catfile( $aligner_dir, $entry );
        next if !-d $path;
        next if  -l $path;

        # Is this really a safe assumption to make?
        if ( $entry =~ m/^ $aligner_name [_-]? (\S+) /msx ) {
          push @list, $1;
        } else {
          next;
        }
    }
    closedir $dir_handle;


    if ( scalar @list == 0 ) {
        carp 'Could not establish the list of installed versions for '
            . $aligner_name;
    }

    return \@list;
}

1;


__END__


=head1 NAME

npg_common::roles::aligners - return information about a central aligner collection

=head1 VERSION

$Revision: 15261 $

=head1 SYNOPSIS

    use Moose;
    with 'npg_common::roles::aligners';

    my $base            = $self->collection_root();
    my @tools           = @{ $self->aligner_list() };
    my $current_bowtie  = $self->current_version('bowtie');
    my @bowtie_versions = @{ $self->installed_versions('bowtie') };

=head1 DESCRIPTION

The NPD maintains a collection of various aligners, adding new versions as
they are released. A symlink called 'current' points to the latest stable
release in each case. See the README file in the collection's base directory
for more information.

Not all of the software maintained in the collection are actually aligners.
Some are better described as alignment tools (e.g. samtools, picard), and some
are assemblers rather than aligners (e.g. fuzzypath). Regardless, throughout
this documentation the term 'aligner' will be understood to cover all
software maintained in the collection.

This role returns useful information about the collection and the aligners
maintained there.

=head1 SUBROUTINES/METHODS

=head2 current_base_dir

Return the path to the directory holding the most recent version of the
aligner executables. It's up to the user to know what these executables are
called.

The path is constructed according to WTSI conventions. It there is no
directory found at the path, the method carps but returns the path found
anyway.

The aligner name is required as an argument to the method.

=head2 current_version

Try to get the latest installed version of an aligner to self-report its
version string. If that fails fall back to name of the directory that the
'current' symlink points to. It that fails return a 'no version found' string.

The aligner name is required as an argument to the method.

=head2 installed_versions

List all the installed versions of an aligner found in the collection. It
reports the name of the subdirectories for each installed version rather than
a version string as in the method 'current_version' because this is more
useful in constructing a path to a non-current version of the aligner.

The aligner name is required as an argument to the method.

=head1 CONFIGURATION AND ENVIRONMENT

The current_version method works by calling the aligner with various arguments
and parsing the output for a version string. Not all aligners have executables
named after the aligner package, and not all of them report the version string
in a uniform way.

There are two hard-coded hashes to deal with this that will need updating for
each newly installed aligner which falls into the above category.

=head1 DEPENDENCIES

=head1 INCOMPATIBILITIES

=head1 DIAGNOSTICS

=head1 BUGS AND LIMITATIONS

Please contact the author with any found.

=head1 AUTHOR

John O'Brien, E<lt>jo3@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by John O'Brien

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
