#########
# Author:        gq1
# Maintainer:    $Author: dj3 $
# Created:       2010 01 15
# Last Modified: $Date: 2012-02-29 13:11:34 +0000 (Wed, 29 Feb 2012) $
# Id:            $Id: BAM_Generation.pm 15261 2012-02-29 13:11:34Z dj3 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-43.0/lib/npg_common/sequence/BAM_Generation.pm $
#

package npg_common::sequence::BAM_Generation;

use strict;
use warnings;
use Moose;
use Carp;
use English qw(-no_match_vars);
use File::Spec::Functions qw(catfile);
use File::Temp qw(tempfile tempdir);
use Cwd qw(cwd abs_path);
use File::Basename;

use st::api::lims;
use npg_common::Alignment;
use npg_common::sequence::reference;
use npg_common::sequence::reference::info;
use npg_common::extractor::fastq qw/first_read_length/;

with qw/
        MooseX::Getopt
        npg_common::roles::run
        npg_common::roles::run::lane
       /;

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 15261 $ =~ /(\d+)/mxs; $r; };

Readonly::Scalar our $DEFAULT_BWA_TRIM_QUALITY => q{-q 15};
Readonly::Scalar our $DEFAULT_BWA_THREADS      => q{-t 2};
Readonly::Scalar our $MINUS_ONE                => -1;
Readonly::Scalar our $MIN_READ_LENGTH_TO_ALIGN => 31;

## no critic (Documentation::RequirePodAtEnd)

=head1 NAME

npg_common::sequence::BAM_Generation

=head1 VERSION

$LastChangedRevision: 15261 $

=head1 SYNOPSIS

  my $bam = npg_common::sequence::BAM_Generation->new({
                 id_run        => $id_run,
                 position      => $position,
                 fastq_files   => [$fastq1, $fastq2],
                 bam_filename  => $output_dir,
               });
  $bam->generate();

=head1 DESCRIPTION


=head1 SUBROUTINES/METHODS

=cut

has '+id_run'   =>        (required        => 0,);
has '+position' =>        (required        => 0,);

=head2 tag_index 	 
	  	 
tag_index if the sam file is for one tag 	 
	  	 
=cut 	 
has 'tag_index'    => (isa           => q{Int},
	               is            => q{rw},
	               required      => 0,
	               documentation => q{tag index},
	              );

=head2 _lims

lims access object

=cut
has  '_lims'    =>  ( isa           => 'st::api::lims',
                      is            => 'ro',
                      init_arg        => undef,
                      lazy_build      => 1,
                    );
sub _build__lims {
    my $self = shift;
    return st::api::lims->new(id_run => $self->id_run, position => $self->position, tag_index => $self->tag_index);
}

=head2 human_split

to indicate this sam file is split or not, and human part or nonhuman part 

=cut
has 'human_split'  => (isa           => q{Str},
                       is            => q{rw},
                       required      => 0,
                       documentation => q{human_split type},
                      );

=head2 archive_path

runfolder archive path

=cut
has 'archive_path' => (isa           => q{Str},
                       is            => q{rw},
                       required      => 0,
                       documentation => q{runfolder archive path},
                      );

=head2 no_alignment

skip the alignment, generate bam file without alignment

=cut
has 'no_alignment' => (isa           => q{Bool},
                       is            => q{rw},
                       required      => 0,
                       lazy_build    => 1,
                       documentation => q{skip alignment},
                      );
sub _build_no_alignment {
  my $self = shift;

  if( $self->human_split() && ( $self->human_split() eq q{phix} ) ) {
     return 0;
  }

  my $no_alignment = $self->_lims->alignments_in_bam ? 0 : 1;
  if (!$no_alignment) {
    my $zero4empty_file = 1;
    foreach my $file (@{$self->fastq_files}) {
      if (first_read_length($file, $zero4empty_file) < $self->min_read_length_to_align) {
	$no_alignment = 1;
	last;
      }
    }
  }

  return $no_alignment;
}


=head2 min_read_length_to_align

If read length is smaller than this number, no_alignment option will be set to true if not set by the caller. The default value is 31.

=cut
has 'min_read_length_to_align' => (isa           => q{Int},
                                   is            => q{ro},
                                   required      => 0,
                                   default       => $MIN_READ_LENGTH_TO_ALIGN,
                                   documentation => q{If read length is smaller than this number, no_alignment option will be set to true if not set by the caller},
                                  );


=head2 reference_info

an object of reference_info including bwa options and reference prefix for BWA alignment

=cut

has 'reference_info'  => (isa           => 'Maybe[npg_common::sequence::reference::info]',
                          is            => 'rw',
                          required      => 0,
                          lazy_build    => 1,
                          documentation => 'Reference info object including bwa options and reference prefix for BWA alignment',
                         );

sub _build_reference_info {
  my $self = shift;

  my $ref_bwa = npg_common::sequence::reference->new(
             lims      => $self->_lims,
             id_run    => $self->id_run(),
             position  => $self->position(),
             tag_index => $self->tag_index(),
            );

  my $ref_info;

  eval {
    $ref_info = $ref_bwa->ref_info();
    1;
  } or do {
    my $error_message = qq{Error: reference bwa options cannot be retrieved; $EVAL_ERROR};
    carp $error_message;
  };

  return $ref_info;
}

=head2 reference

A path to reference

=cut

has 'reference'   => (isa           => 'Str',
                      is            => 'rw',
                      required      => 0,
                      lazy_build    => 1,
                      documentation => 'Reference prefix for BWA alignment',
                     );

sub _build_reference {
  my $self = shift;

  my $ref_info = $self->reference_info();

  if(!$ref_info){
     return q{};
  }

  my $reference = $ref_info->ref_path();
  my($filename, $directories, $suffix) = fileparse($reference);

  return catfile(abs_path($directories), $filename);
}

=head2 reference_fai

A path to reference in fai

=cut

has 'reference_fai'=> (isa           => 'Str',
                       is            => 'rw',
                       required      => 0,
                       lazy_build    => 1,
                       documentation => 'Reference location for samtools (fai) with fasta in the same location',
                     );

sub _build_reference_fai {
  my $self = shift;

  my $ref_fasta = npg_common::sequence::reference->new(
             lims      => $self->_lims,
             id_run => $self->id_run(),
             position  => $self->position(),
             aligner   => q{fasta},
             tag_index => $self->tag_index(),
            );
  my $refs;
  eval {
    $refs = $ref_fasta->refs();
    1;
  } or do {
    my $error_message = qq{Error: binary reference fasta cannot be retrieved; $EVAL_ERROR};
    carp $error_message;
    return q{};
  };

  if (scalar @{$refs} > 1) {
    carp q[Run ] . $self->id_run . q[ lane ] . $self->position . q[ - multiple references for a pooled lane: ] . (join q[;], @{$refs});
    return q{};
  }

  if(! scalar @{$refs}){
    return q{};
  }

  my $reference_fa = pop @{$refs};

  my $reference_fai = $reference_fa.q{.fai};

  if(! -e $reference_fai){
    croak q[Run ] . $self->id_run . q[ lane ] . $self->position . q[ - reference fai missing; ].$reference_fai;
  }
  return $reference_fai;
}

=head2 aligner_cmd

bwa command 

=cut

has 'aligner_cmd'     => (isa            => 'Str',
                          is             => 'rw',
                          required       => 0,
                          default        => abs_path('/software/solexa/bin/bwa'),
                          trigger        => \&_covert_aligner_path,
                          documentation  => 'aligner command, default bwa',
                         );

sub _covert_aligner_path {
   my ($self) = @_;

   my $old_path =  $self->aligner_cmd();
   my $abs_path = abs_path $old_path;

   if($abs_path ne $old_path){
      $self->aligner_cmd($abs_path);
   }

   return;
}

=head2 aligner_options

bwa command options
 
=cut

has 'aligner_options' => (isa             => 'Str',
                          is              => 'rw',
                          required        => 0,
                          lazy_build      => 1,
                          documentation   => 'aligner command options',
                         );

sub _build_aligner_options {
  my $self = shift;

  my $ref_info = $self->reference_info();

  my $aligner_options = $DEFAULT_BWA_TRIM_QUALITY;;

  if( $ref_info ){

     $aligner_options = $ref_info->aligner_options();
  }

  if( $aligner_options eq $npg_common::sequence::reference::roles::find::NPG_DEFAULT_ALIGNER_OPTION ){
      $aligner_options = $DEFAULT_BWA_TRIM_QUALITY;
  }

  return $aligner_options.q{ }.$DEFAULT_BWA_THREADS;
}

=head2 samtools_cmd

samtools command 

=cut

has 'samtools_cmd'     => (isa           => 'Str',
                          is             => 'rw',
                          required       => 0,
                          default        =>
                            sub { abs_path '/software/solexa/bin/aligners/samtools/current/samtools'; },
                          trigger        => \&_covert_samtools_path,
                          documentation  => 'samtools command',
                         );

sub _covert_samtools_path {
   my ($self) = @_;

   my $old_path =  $self->samtools_cmd();
   my $abs_path = abs_path $old_path;

   if($abs_path ne $old_path){
      $self->samtools_cmd($abs_path);
   }

   return;
}

=head2 picard_converter_cmd

Picard  SamFormatConverter command take input from stdin

=cut

has 'picard_converter_cmd'       => ( isa             => 'Str',
                                      is              => 'rw',
                                      default         => q{/software/bin/java -jar }.abs_path(q{/software/solexa/bin/aligners/picard/current/SamFormatConverter.jar}).q{ VALIDATION_STRINGENCY=SILENT INPUT=/dev/stdin},
                                      trigger        => \&_covert_picard_path,
                                      documentation   => 'Picard SamFormatCoverter command',
                        );

sub _covert_picard_path {
   my ($self) = @_;

   my $old_path =  $self->picard_converter_cmd();
   my @fields =  map { $_ =~ /picard/mxs  ? abs_path($_): $_  } split  /\s/mxs, $old_path;
   my $abs_path = join q{ },  @fields;

   if($abs_path ne $old_path){
      $self->picard_converter_cmd($abs_path);
   }

   return;
}

=head2 sam_second_call_cmd

add second call to sam command

=cut

has 'sam_second_call_cmd' => (isa           => 'Str',
                              is             => 'rw',
                              required       => 0,
                              default        => '/software/solexa/bin/sam_second_call.pl',
                              documentation  => 'add second call to sam command',
                            );

=head2 second_call_qseq_dir

second call qseq directory
 
=cut

has 'second_call_qseq_dir'=> ( isa             => 'Str',
                               is              => 'rw',
                               required        => 0,
                               documentation   => 'second call qseq dir',
                             );
=head2 read_number_1

read number 1

=cut

has 'read_number_1'            => ( isa            => 'Int',
                                    is             => 'rw',
                                    documentation  => 'read number 1',
                                  );
=head2 read_number_2

read number 2

=cut

has 'read_number_2'            => ( isa            => 'Int',
                                    is             => 'rw',
                                    documentation  => 'read number 2',
                                  );

=head2 sam_original_quality_cmd

add original quality to sam as OQ tag command

=cut

has 'sam_original_quality_cmd' => (isa            => 'Str',
                                   is             => 'rw',
                                   required       => 0,
                                   default        => '/software/solexa/bin/sam_original_quality.pl',
                                   documentation  => 'add original quality to sam as OQ tag command',
                                  );
=head2 original_quality_qseq_dir

original quality qseq file directory
 
=cut

has 'original_quality_qseq_dir'=> ( isa             => 'Str',
                                    is              => 'rw',
                                    required        => 0,
                                    documentation   => 'original quality qseq file directory',
                                  );
=head2 sam_fastq_check_cmd

sam_fastq_check command

=cut

has 'sam_fastq_check_cmd' => (isa            => 'Str',
                              is             => 'rw',
                              required       => 0,
                              default        => '/software/solexa/bin/sam_fastq_check.pl',
                              documentation  => 'sam_fastq_check command',
                             );

=head2 fasq2sam_cmd

fastq2sam command to genrerate sam file when no alignment available

=cut

has 'fastq2sam_cmd'   => (isa           => 'Str',
                          is             => 'rw',
                          required       => 0,
                          default        => '/software/solexa/bin/fastq2sam.pl',
                          documentation  => 'fastq to sam command to generate sam file without alignment',
                         );
=head2 sam_header_cmd

sam_header_cmd to add some header sections to a sam file

=cut

has 'sam_header_cmd'   => (isa           => 'Str',
                          is             => 'rw',
                          required       => 0,
                          default        => '/software/solexa/bin/sam_header.pl',
                          documentation  => 'sam_header_cmd to add some header sections to a sam file',
                         );
=head2 fastq_files

one or two fastq file names
 
=cut

has 'fastq_files'    => ( isa             => 'ArrayRef',
                          is              => 'rw',
                          required        => 1,
                          documentation   => 'fasta files in array',
                        );

=head2 sam_index_tag_cmd

add index tag to sam command

=cut

has 'sam_index_tag_cmd'=> (isa           => 'Str',
                          is             => 'rw',
                          required       => 0,
                          default        => '/software/solexa/bin/sam_index_tag.pl',
                          documentation  => 'add index tag to sam command',
                         );
=head2 index_fastq_file

index_fastq file to add index tag to sam file
 
=cut

has 'index_fastq_file'=> ( isa            => 'Str',
                          is              => 'rw',
                          required        => 0,
                          documentation   => 'index_fastq_file',
                        );

=head2 bam_filename

output directory
 
=cut

has 'bam_filename'   => ( isa             => 'Str',
                          is              => 'rw',
                          required        => 1,
                          documentation   => 'output bam filename with path',
                        );

=head2 generate

- use given fastq, doing alignment and generate bam file in output directory
 
=cut

sub generate {
  my $self = shift;

  my $reference;
  my $reference_fai;

  #get reference if alignment required
  if( !$self->no_alignment() ) {
    $reference = $self->reference();
    $reference_fai = $self->reference_fai();
  }

  #check fastq files
  my @fastq_files = @{$self->fastq_files};

  foreach my $fastq (@fastq_files){
    if( !-e $fastq ){
      croak "Fastq file $fastq does not exist, no bam will be generated";
    }
  }

  #only generate bam header if fastq file size zero
  foreach my $fastq (@fastq_files){
    if( -s $fastq == 0 ){
      carp "Fastq file $fastq size is zero, bam will be generated only with header";
      $self->_generate_bam_no_alignment();
      return 1;
    }
  }

  #generate bam without alignment if no reference found or no alignment required
  if(!$reference || !$reference_fai){
    carp 'reference not available, bam file will be generated without alignment!';
    $self->_generate_bam_no_alignment();
    return 1;
  }


  #bam file in output directory, without extension .bam
  my $bam_filename = $self->bam_filename();

  #temp unsorted file name
  my $work_dir = tempdir( CLEANUP => 1 );

  #doing alignment
  my $bwa = npg_common::Alignment->new(
               {
                 bwa_cmd              => $self->aligner_cmd(),
                 bwa_options          => $self->aligner_options(),
                 sam_fastq_check_cmd  => $self->sam_fastq_check_cmd,
                 sam_second_call_cmd  => $self->sam_second_call_cmd(),
                 sam_original_quality_cmd  => $self->sam_original_quality_cmd(),
                 sam_index_tag_cmd    => $self->sam_index_tag_cmd(),
                 sam_header_cmd       => $self->sam_header_cmd(),
                 picard_converter_cmd => $self->picard_converter_cmd().q{ COMPRESSION_LEVEL=0 OUTPUT=/dev/stdout QUIET=true},
                 samtools_cmd         => $self->samtools_cmd(),
               });

  if($self->original_quality_qseq_dir()){
    $bwa->sam_original_quality_options($self->_original_quality_options());
  }

  if($self->second_call_qseq_dir()){
    $bwa->sam_second_call_options($self->_second_call_options());
  }

  my $align_args_ref = { bam_out          => $bam_filename,
                         ref_root         => $reference,
                         ref_fai          => $reference_fai,
                         id_run           => $self->id_run(),
                         position         => $self->position(),
                         human_split      => $self->human_split(),
                         tag_index        => $self->tag_index(),
                         index_fastq_file => $self->index_fastq_file(),
                         archive_path     => $self->archive_path(),
                       };

  #generate bam file, unsorted
  if(scalar @fastq_files == 1 ){

    $align_args_ref->{fastq} = $fastq_files[0];

    $bwa->bwa_align_bam_se( $align_args_ref );
  }elsif(scalar @fastq_files == 2 ){

    $align_args_ref->{fastq1}     = $fastq_files[0];
    $align_args_ref->{fastq2}     = $fastq_files[1];
    $align_args_ref->{fork_align} = 1;

    $bwa->bwa_align_bam_pe( $align_args_ref );
  }else{
    croak "There are more than two fastq given:\n".( join "\n", @fastq_files );
  }

  return 1;
}

sub _generate_bam_no_alignment{
  my $self = shift;

  #check fastq files
  my $num_fastq = scalar @{$self->fastq_files()};
  my ($fastq1, $fastq2);
  if( $num_fastq == 1 ){
    $fastq1 = $self->fastq_files()->[0];
  }elsif($num_fastq == 2){
    $fastq1 = $self->fastq_files()->[0];
    $fastq2 = $self->fastq_files()->[1];
  }else{
     croak 'no fastq or more than 2 fastq file available';
  }

  #output bam file in output directory
  my $bam_filename = $self->bam_filename().q{.bam};

  #fastq2sam
  my $fastq2bam_cmd;
  my $fastq2sam_cmd = $self->fastq2sam_cmd().qq{ --fastq1 $fastq1};
  if($fastq2){
    $fastq2sam_cmd .= qq{ --fastq2 $fastq2};
  }
  $fastq2bam_cmd = $fastq2sam_cmd;

  #add original quality
  my $add_original_quality;
  if($self->original_quality_qseq_dir()){
    $add_original_quality = $self->sam_original_quality_cmd.q{ }.$self->_original_quality_options();
    $fastq2bam_cmd .= q{ | }.$add_original_quality;
  }

  #add second call
  my $add_second_call;
  if($self->second_call_qseq_dir()){
    $add_second_call = $self->sam_second_call_cmd.q{ }.$self->_second_call_options();
    $fastq2bam_cmd .= q{ | }.$add_second_call;
  }

  #add index sequences
  my $add_index_fastq;
  if( $self->index_fastq_file() ){
     $add_index_fastq = $self->sam_index_tag_cmd . q{ --index_fastq_file } . $self->index_fastq_file();
     $fastq2bam_cmd .= q{ | }.$add_index_fastq
  }

  my $picard_converter_cmd = $self->picard_converter_cmd().qq{ OUTPUT=$bam_filename};

  #add header
  if($self->sam_header_cmd()){
     my $add_header = $self->sam_header_cmd().q{ --id_run }. $self->id_run();

     my $position = $self->position();
     if($position){
       $add_header .= qq{ --position $position};
     }

     if( $self->archive_path() ){
       $add_header .= q{ --archive_path }.$self->archive_path();
     }

     my $human_split = $self->human_split;
     if($human_split){
       $add_header .= qq{ --human_split $human_split};
     }

     my $tag_index = $self->tag_index();
     if(defined $tag_index){
       $add_header .= qq{ --tag_index $tag_index};
     }

     $add_header .= q{ --fastq2sam_cmd '}.$fastq2sam_cmd.q{'};

     if( $self->index_fastq_file() ){
       $add_header .= q{ --sam_index_tag_cmd '}.$add_index_fastq.q{'};
     }

     if( $self->original_quality_qseq_dir() ){
       $add_header .= q{ --sam_original_quality_cmd '}.$add_original_quality.q{'};
     }

     if( $self->second_call_qseq_dir() ){
       $add_header .= q{ --sam_second_call_cmd '}.$add_second_call.q{'};
     }

     $add_header .= q{ --no_alignment};

     $add_header .= q{ --picard_converter_cmd '}.$picard_converter_cmd.q{'};

     $fastq2bam_cmd .= q{ | }.$add_header;
  }

  #covert sam to bam
  $fastq2bam_cmd  .= q{ | }.$picard_converter_cmd;

  #run command
  print {*STDERR} "Covert fastq files to bam without alignment:$fastq2bam_cmd\n" or croak $ERRNO;

  my $fastq2bam_rs = system "set -o pipefail; $fastq2bam_cmd";
  if($fastq2bam_rs != 0){
    croak 'Fastq to bam without alignment failed!!!';
  }

  return 1;
}

sub _second_call_options{
  my $self = shift;

  my $second_call_options = q{--lane }.$self->position().q{ --qseq_dir }.$self->second_call_qseq_dir();
  if($self->read_number_1()){
       $second_call_options .= q{ --read_number_1 }.$self->read_number_1();
  }
  if($self->read_number_2()){
       $second_call_options .= q{ --read_number_2 }.$self->read_number_2();
  }
  return $second_call_options;
}

sub _original_quality_options{
  my $self = shift;

  my $original_quality_options = q{--lane }.$self->position().q{ --qseq_dir }.$self->original_quality_qseq_dir();

  # read numbers for these qseq files has been renamed
  # for single read run, it always is 1
  # for paired read run, it always is 1 and 2
  # for indexing read, it always is t

  # the above is not correct now

  # read numbers now have been changed, using the ones from command line
  if($self->read_number_1()){
    $original_quality_options .= q{ --read_number_1 }.$self->read_number_1();
  }
  if($self->read_number_2()){
    $original_quality_options .= q{ --read_number_2 }.$self->read_number_2();
  }

  return $original_quality_options;
}


no Moose;

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item MooseX::Getopt

=item File::Spec::Functions

=item File::Temp

=item Cwd

=item File::Basename

=item Carp

=item English -no_match_vars

=item Readonly

=item npg_common::sequence::reference

=item st::api::lims

=item npg_common::Alignment

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Guoying Qi E<lt>gq1@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2009 GRL, by Guoying Qi

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
