#########
# Author:        gq1
# Maintainer:    $Author: gq1 $
# Created:       2010 06 21
# Last Modified: $Date: 2012-02-02 10:54:00 +0000 (Thu, 02 Feb 2012) $
# Id:            $Id: BAM_MarkDuplicate.pm 15093 2012-02-02 10:54:00Z gq1 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-43.0/lib/npg_common/sequence/BAM_MarkDuplicate.pm $
#

package npg_common::sequence::BAM_MarkDuplicate;

use strict;
use warnings;
use Moose;
use Carp;
use English qw(-no_match_vars);
use File::Temp qw/ tempfile tempdir /;
use Perl6::Slurp;
use IPC::Open3;
use npg_qc::autoqc::results::bam_flagstats;
use npg_qc::autoqc::checks::check;
use Cwd qw(abs_path);
use File::Basename;
use autodie qw(:all);

with qw/
        MooseX::Getopt
        npg_common::roles::log
       /;

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 15093 $ =~ /(\d+)/mxs; $r; };

Readonly::Scalar our $DEFAULT_JAVA_CMD               => q{/software/bin/java};
Readonly::Scalar our $DEFAULT_MARK_DUPLICATES_JAR    => q{/software/solexa/bin/aligners/picard/current/MarkDuplicates.jar};
Readonly::Scalar our $DEFAULT_JAVA_XMX_DUPLICATES    => 3000;
Readonly::Scalar our $DEFAULT_COMMAND_OPTIONS        => {
                               VALIDATION_STRINGENCY => 'SILENT',
                               VERBOSITY             => 'ERROR',
                               ASSUME_SORTED         => 'true',
                               READ_NAME_REGEX       => '[a-zA-Z0-9_]+:[0-9]:([0-9]+):([0-9]+):([0-9]+).*',
                               REMOVE_DUPLICATES     => 'false',
                               CREATE_MD5_FILE       => 'TRUE',
                               CREATE_INDEX          => 'TRUE',
                               MAX_FILE_HANDLES_FOR_READ_ENDS_MAP => 900,
                                                 };

Readonly::Scalar our $DEFAULT_SAMTOOLS_CMD           => q{/software/solexa/bin/aligners/samtools/current/samtools};

Readonly::Scalar our $ILLUMINA2BAM_DIR               => q{/software/solexa/bin/aligners/illumina2bam/current};
Readonly::Scalar our $DEFAULT_CHANGE_BAM_HEADER_JAR  => $ILLUMINA2BAM_DIR . q{/ChangeBamHeader.jar};
Readonly::Scalar our $DEFAULT_BAM_TAG_STRIPPER_JAR   => $ILLUMINA2BAM_DIR . q{/BamTagStripper.jar};
Readonly::Scalar our $DEFAULT_JAVA_XMX_ILLUMINA2BAM  => 1000;
Readonly::Scalar our $DEFAULT_ILLUMINA2BAM_OPTIONS   => {
                               VALIDATION_STRINGENCY => 'SILENT',
                               VERBOSITY             => 'INFO',
                               CREATE_MD5_FILE       => 'TRUE',
                               CREATE_INDEX          => 'TRUE',
                                                 };
Readonly::Scalar our $BAM_TAGS_TO_STRIP              => [qw(OQ)];

## no critic (Documentation::RequirePodAtEnd)

=head1 NAME

npg_common::sequence::BAM_MarkDuplicate

=head1 VERSION

$LastChangedRevision: 15093 $

=head1 SYNOPSIS

  my $bam = npg_common::sequence::BAM_MarkDuplicate->new({
                 input_bam     => 'input.bam',
                 output_bam    => 'output.bam',
                 metrics_json  => 'metrics.json',

               });
  $bam->process();

=head1 DESCRIPTION


=head1 SUBROUTINES/METHODS

=cut

=head2 java_command

JAVA command

=cut
has 'java_cmd'       => ( isa             => 'Str',
                          is              => 'rw',
                          default         => $DEFAULT_JAVA_CMD,
                          documentation   => 'java command',
                        );

=head2 java_Xmx

JAVA command Xmx value

=cut
has 'java_Xmx'       => ( isa             => 'Int',
                          is              => 'rw',
                          default         => $DEFAULT_JAVA_XMX_DUPLICATES,
                          documentation   => 'java command Xmx value',
                        );

=head2 jar_file

Picard MarkDuplicates jar file

=cut

has 'jar_file'       => ( isa             => 'Str',
                          is              => 'rw',
                          default         => abs_path($DEFAULT_MARK_DUPLICATES_JAR),
                          trigger         => \&_covert_picard_path,
                          documentation   => 'Picard MarkDuplicates jar file',
                        );

sub _covert_picard_path {
   my ($self) = @_;

   my $old_path =  $self->jar_file();
   my $abs_path = abs_path($old_path);

   if($abs_path ne $old_path){
      $self->jar_file($abs_path);
   }
   return;
}

=head2 mark_duplicate_options

=cut

has 'mark_duplicate_options'       => ( isa             => 'HashRef',
                                        is              => 'rw',
                                        default         => sub { $DEFAULT_COMMAND_OPTIONS },
                                        documentation   => 'Picard MarkDuplicates command options',
                                      );

=head2 change_bam_header_jar_file

illumina2bam ChangeBamHeader jar file

=cut

has 'change_bam_header_jar_file' => ( isa             => 'Str',
                                      is              => 'rw',
                                      default         => abs_path($DEFAULT_CHANGE_BAM_HEADER_JAR),
                                      trigger         => \&_covert_change_bam_header_path,
                                      documentation   => 'illumina2bam ChangeBamHeader jar file',
                                    );

sub _covert_change_bam_header_path {
   my ($self) = @_;

   my $old_path =  $self->change_bam_header_jar_file();
   my $abs_path = abs_path($old_path);

   if($abs_path ne $old_path){
      $self->change_bam_header_jar_file($abs_path);
   }
   return;
}

=head2 bam_tag_stripper_jar_file

illumina2bam BamTagStripper jar file

=cut

has 'bam_tag_stripper_jar_file'  => ( isa             => 'Str',
                                      is              => 'rw',
                                      default         => abs_path($DEFAULT_BAM_TAG_STRIPPER_JAR),
                                      trigger         => \&_covert_bam_tag_stripper_path,
                                      documentation   => 'illumina2bam BamTagStripper jar file',
                                    );

sub _covert_bam_tag_stripper_path {
   my ($self) = @_;

   my $old_path =  $self-> bam_tag_stripper_jar_file();
   my $abs_path = abs_path($old_path);

   if($abs_path ne $old_path){
      $self->bam_tag_stripper_jar_file($abs_path);
   }
   return;
}

=head2 samtools_cmd

samtools command

=cut
has 'samtools_cmd'        => ( isa             => 'Str',
                               is              => 'ro',
                               default         => abs_path($DEFAULT_SAMTOOLS_CMD),
                               documentation   => 'samtools command',
                             );

=head2 input_bam

input bam file name with path

=cut

has 'input_bam'      => ( isa             => 'Str',
                          is              => 'rw',
                          required        => 1,
                          documentation   => 'input bam filename with path',
                        );

=head2 sort_input

sort input bam file by coordinates

=cut
has 'sort_input'  => (isa             => 'Bool',
                      is              => 'rw',
                      required        => 0,
                      documentation   => 'sort input bam file by coordinates',
                     );

=head2 change_bam_header

add extra PG records into output header

=cut
has 'change_bam_header'  => (isa             => 'Bool',
                             is              => 'rw',
                             required        => 0,
                             documentation   => 'add sorting or markduplicates steps as PG records into output header',
                           );

=head2 not_strip_bam_tag

not strip any tag from output bam records

=cut
has 'not_strip_bam_tag'  => (isa             => 'Bool',
                             is              => 'rw',
                             required        => 0,
                             documentation   => 'not strip any tag from output bam records',
                           );

=head2 output_bam

output bam with path

=cut

has 'output_bam'     => ( isa             => 'Str',
                          is              => 'rw',
                          required        => 1,
                          documentation   => 'output bam filename with path',
                        );
=head2 metrics_file

output metrics file name

=cut

has 'metrics_file'   => ( isa             => 'Str',
                          is              => 'rw',
                          required        => 0,
                          lazy_build      => 1,
                          documentation   => 'output metrics file name',
                        );

sub _build_metrics_file {

   my ($fh, $filename) = tempfile(UNLINK => 1);
   return $filename;
}

=head2 temp_dir

temp dir

=cut

has 'temp_dir'       => ( isa             => 'Str',
                          is              => 'rw',
                          lazy_build      => 1,
                          documentation   => 'temp dir',
                        );
sub _build_temp_dir {
   my $self = shift;
   #return tempdir( CLEANUP => 1 );
   my $output = $self->output_bam();
   my ($name,$path,$suffix) = fileparse($output);
   return tempdir(CLEANUP => 1, DIR => $path);
}

=head2 sorted_input_bam_prefix

sorted input bam prefix with path

=cut

has 'sorted_input_bam_prefix'=> ( isa            => 'Str',
                                  is             => 'rw',
                                  lazy_build     => 1,
                                  documentation  => 'sorted input bam prefix with path',
                                );

sub _build_sorted_input_bam_prefix {
   my $self = shift;

   return $self->temp_dir().q{/sorted};
}

=head2 _result

autoqc result object

=cut
has '_result'              => ( isa            => 'npg_qc::autoqc::results::bam_flagstats',
                               is              => 'rw',
                               lazy_build      => 1,
                               documentation   => 'autoqc result object',
                             );
sub _build__result {
   my $self = shift;
   my $result= npg_qc::autoqc::results::bam_flagstats->new();
   if($self->id_run()){
      $result->id_run($self->id_run);
   }
   if($self->position()){
      $result->position($self->position);
   }
   if(defined $self->tag_index){
      $result->tag_index($self->tag_index);
   }
   if($self->human_split){
      $result->human_split($self->human_split);
   }
   return $result;
}

=head2 metrics_json

metrics json output file name

=cut
has 'metrics_json'   => ( isa             => 'Str',
                          is              => 'rw',
                          required        => 1,
                          documentation   => 'output json metrics file name',
                       );

=head2 id_run

Run id for the lane to be checked.

=cut
has 'id_run'      => (
                       isa            => 'Int',
                       is             => 'rw',
                       required       => 0,
                       documentation  => 'only for metrics json output',
		      );

=head2 position

Lane number. An integer from 1 to 8 inclusive.

=cut
has 'position'    => (isa       => 'Int',
                      is        => 'rw',
                      required  => 0,
                      documentation  => 'only for metrics json output',
                     );

=head2 tag_index

Tag index. An integer from 0 to 10000. Zero means that no tag has been matched.

=cut
has 'tag_index'   => (isa        => 'Int',
                      is         => 'rw',
                      required   => 0,
                      documentation  => 'only for metrics json output',
                     );

=head2 human_split

to show the bam file is phix, human part or nonhuman part
default 'all', which means not split

=cut
has 'human_split'   => (isa        => 'Str',
                        is         => 'rw',
                        required   => 0,
                        documentation  => 'only for metrics json output',
                       );

=head2 replace_file

replace the input bam file with the output duplicates marked bam file

=cut
has 'replace_file'=> (isa        => 'Bool',
                      is         => 'rw',
                      required   => 0,
                      documentation   => 'replace the input bam file with the output duplicates marked bam file',
                     );

=head2 mark_duplicate_cmd

construct the whole picard markduplicate command

=cut
has 'mark_duplicate_cmd'  => ( isa             => 'Str',
                               is              => 'ro',
                               lazy_build      => 1,
                               documentation   => 'read only mark duplicate command',
                             );

sub _build_mark_duplicate_cmd {
   my $self = shift;

   my $cmd = $self->java_cmd.q{ -Xmx}.$self->java_Xmx().q{m};
   $cmd .= q{ -jar }.$self->jar_file();

   if( $self->sort_input() ){
      $cmd .= q{ INPUT=} . $self->sorted_input_bam_prefix() . q{.bam};
   }else {
      $cmd .= q{ INPUT=}.$self->input_bam();
   }
   my %options = %{$self->mark_duplicate_options()};
   if( $self->change_bam_header() || !$self->not_strip_bam_tag() ){
      $cmd .= q{ OUTPUT=/dev/stdout};
      $options{COMPRESSION_LEVEL} = 0;
      delete $options{CREATE_MD5_FILE};
      delete $options{CREATE_INDEX};
   }else{
      $cmd .= q{ OUTPUT=}.$self->output_bam();
   }

   $cmd .= q{ METRICS_FILE=}.$self->metrics_file();
   #$cmd .= q{ TMP_DIR=}.$self->temp_dir();

   foreach my $option_key (sort keys %options){
       $cmd .= qq{ $option_key='}.$options{$option_key}.q{'};
   }
   return $cmd;
}


=head2 _mark_duplicate_pg

construct the whole picard markduplicate PG for output bam header

=cut
has '_mark_duplicate_pg'  => ( isa             => 'Str',
                               is              => 'ro',
                               lazy_build      => 1,
                               documentation   => 'construct the whole picard markduplicate PG for output bam header',
                             );

sub _build__mark_duplicate_pg {
   my $self = shift;
   my $pg = q{ID:picard_markduplicates;PN:Picard;VN:}
           . $self->_get_picard_version()
           . q{;CL:} . $self->mark_duplicate_cmd();
   return $pg;
}

=head2 change_bam_header_cmd

construct the whole illumina2bam ChangeBamHeader command

=cut
has 'change_bam_header_cmd'  => ( isa             => 'Str',
                               is              => 'ro',
                               lazy_build      => 1,
                               documentation   => 'construct the whole illumina2bam ChangeBamHeader command',
                             );

sub _build_change_bam_header_cmd {
   my $self = shift;

   my %options = %{$DEFAULT_ILLUMINA2BAM_OPTIONS};

   my $cmd = $self->java_cmd.q{ -Xmx}.$DEFAULT_JAVA_XMX_ILLUMINA2BAM.q{m};
   $cmd .= q{ -jar }.$self->change_bam_header_jar_file();
   $cmd .= q{ INPUT=/dev/stdin};

   if(!$self->not_strip_bam_tag()){
      $cmd .= q{ OUTPUT=/dev/stdout};
      delete $options{CREATE_MD5_FILE};
      delete $options{CREATE_INDEX};
      $options{COMPRESSION_LEVEL} = 0;
   }else{
      $cmd .= q{ OUTPUT=}.$self->output_bam();
   }

   $cmd .= q{ TMP_DIR=}.$self->temp_dir();

   foreach my $option_key (sort keys %options){
       $cmd .= qq{ $option_key='}.$options{$option_key}.q{'};
   }

   if( $self->sort_input() ) {
      $cmd .= q{ PG="}. $self->_samtools_sort_pg() .q{"};
   }
   $cmd .= q{ PG="}. $self->_mark_duplicate_pg() .q{"};

   return $cmd;
}


=head2 bam_tag_stripper_cmd

construct the whole illumina2bam BamTagStripper command

=cut
has 'bam_tag_stripper_cmd'=> ( isa             => 'Str',
                               is              => 'ro',
                               lazy_build      => 1,
                               documentation   => 'construct the whole illumina2bam BamTagStripper command',
                             );

sub _build_bam_tag_stripper_cmd {
   my $self = shift;

   my $cmd = $self->java_cmd.q{ -Xmx}.$DEFAULT_JAVA_XMX_ILLUMINA2BAM.q{m};
   $cmd .= q{ -jar }.$self->bam_tag_stripper_jar_file();
   if( $self->no_alignment() ){
      $cmd .= q{ INPUT=} . $self->input_bam();
   }else{
      $cmd .= q{ INPUT=/dev/stdin};
   }
   $cmd .= q{ OUTPUT=}.$self->output_bam();
   $cmd .= q{ TMP_DIR=}.$self->temp_dir();

   my $options = $DEFAULT_ILLUMINA2BAM_OPTIONS;
   foreach my $option_key (sort keys %{$options}){
       $cmd .= qq{ $option_key='}.$options->{$option_key}.q{'};
   }

   foreach my $tag (sort @{$BAM_TAGS_TO_STRIP}){
      $cmd .= qq{ STRIP='$tag'};
   }

   return $cmd;
}

=head2 samtools_sort_cmd

samtools sort command for the input bam

=cut
has 'samtools_sort_cmd'   => ( isa             => 'Str',
                               is              => 'ro',
                               lazy_build      => 1,
                               documentation   => 'samtools sort command for the input bam',
                             );

sub _build_samtools_sort_cmd {
   my $self = shift;

   my $cmd = $self->samtools_cmd() . q{ sort } . $self->input_bam() . q{ } . $self->sorted_input_bam_prefix();

   return $cmd;
}

=head2 _samtools_sort_pg

samtools sort command for bam header PG

=cut
has '_samtools_sort_pg'   => ( isa             => 'Str',
                               is              => 'ro',
                               lazy_build      => 1,
                               documentation   => 'samtools sort command for bam header PG',
                             );

sub _build__samtools_sort_pg {
   my $self = shift;
   my $pg = q{ID:samtools_sort;PN:samtools;VN:}
           . npg_qc::autoqc::checks::check->aligner_version($self->samtools_cmd())
           . q{;CL:} . $self->samtools_sort_cmd();
   return $pg;
}

=head2 no_alignment

to indicate input bam with alignment or not

=cut
has 'no_alignment'=> (isa             => 'Bool',
                      is              => 'rw',
                      required        => 0,
                      lazy_build      => 1,
                      documentation   => 'indicate input bam with alignment or not',
                     );
sub _build_no_alignment {
   my $self = shift;

   my $no_alignment = 1;

   my $samtools_view_cmd = $self->samtools_cmd() . q{ view -H } . $self->input_bam();

   $self->log($samtools_view_cmd);

   my $pid = open3( undef, my $input_header_fh, undef, $samtools_view_cmd );

   while( my $line = <$input_header_fh>){

      if($line =~ /^\@SQ\t/mxs){
         $no_alignment = 0;
         $self->log('SQ tag found in input bam file, the input bam with alignment: ' . $self->input_bam());
         last;
      }
   }

   close $input_header_fh;

   if( $no_alignment ){
      $self->log('NO SQ tag found in input bam file, the input bam without alignment:' . $self->input_bam());
   }

   return $no_alignment;
}

=head2 process

main method to call

=cut

sub process {
  my $self = shift;

  if( ! -e $self->input_bam() ){
     croak 'Input bam file does not exist to mark duplicate: '.$self->input_bam();
  }

  $self->_version_info();

  if( $self->sort_input() && !$self->no_alignment() ){

      $self->log('Sort input bam file');
      my $samtools_sort_cmd = $self->samtools_sort_cmd();
      my $samtools_sort_rs = system $samtools_sort_cmd;
      if( $samtools_sort_rs ) {
         croak 'Samtools sort failed!';
      }
  }

  my $mark_duplicate_cmd;

  if( $self->no_alignment() ){
     $self->log('The input file does not include any alignment, generate bam flag stats now: '
                . $self->input_bam()
               );
     $self->_flagstats( $self->input_bam() );
     $self->_result->store($self->metrics_json);

     if( ! $self->not_strip_bam_tag() ){
         $mark_duplicate_cmd = $self->bam_tag_stripper_cmd();
     }

  }else{
     $mark_duplicate_cmd = $self->mark_duplicate_cmd();
     my $using_pipe;
     if( $self->change_bam_header() ){
         $mark_duplicate_cmd .= q{ | } . $self->change_bam_header_cmd();
         $using_pipe++;
     }
     if( ! $self->not_strip_bam_tag() ){
         $mark_duplicate_cmd .= q{ | } . $self->bam_tag_stripper_cmd();
         $using_pipe++;
     }
     if($using_pipe){
        $mark_duplicate_cmd = qq{set -o pipefail;$mark_duplicate_cmd};
     }
  }

  if(!$mark_duplicate_cmd){
      return 1;
  }

  $self->log( $mark_duplicate_cmd );
  my $mark_duplicate_rs = system $mark_duplicate_cmd;
  if( $mark_duplicate_rs != 0){
    croak 'Picard MarkDuplicates failed!';
  }

  if(!$self->no_alignment()){

     $self->log('Generating bam flagstats:' . $self->metrics_json);
     $self->_result->parsing_metrics_file($self->metrics_file());
     $self->_flagstats();
     $self->_result->store($self->metrics_json);
  }

  if($self->replace_file){

    $self->log('Replacing input files with duplicates marked file');

    $self->_move_file($self->output_bam, $self->input_bam);

    my $md5_file_name_mk = $self->output_bam.q{.md5};
    my $md5_file_name     = $self->input_bam.q{.md5};

    if( -e $md5_file_name_mk ) {
       $self->_move_file($md5_file_name_mk, $md5_file_name);
    }else{
       $self->log("BAM MD5 file not generated: $md5_file_name");
    }

    my $index_file_name_mk = $self->output_bam;
    my $index_file_name    = $self->input_bam;
    $index_file_name_mk =~ s/bam$/bai/mxs;
    $index_file_name =~ s/bam$/bai/mxs;

    if( -e $index_file_name_mk ) {
       $self->_move_file($index_file_name_mk, $index_file_name);
    }else{
       $self->log("BAM index file not generated: $index_file_name");
    }

  }

  $self->log('Finished in BAM_Markduplicate!');

  return 1;
}

sub _move_file {
   my ($self, $source, $des) = @_;

   my $mv_cmd = q{mv }.$source.q{ }.$des;
   $self->log("$mv_cmd");
   my $mv_cmd_rs = system $mv_cmd;
   if ( $mv_cmd_rs != 0 ){
      croak "Failed: $mv_cmd";
   }
   return;
}

sub _flagstats {
  my ($self, $bam) = @_;

  $bam ||= $self->output_bam();

  my $samtools_cmd = $self->samtools_cmd().q{ flagstat } . $bam;

  $self->log("$samtools_cmd");

  my $pid = open3( undef, my $samtools_output, undef, $samtools_cmd );

  $self->_result->parsing_flagstats($samtools_output);

  close $samtools_output or carp "Failed to close $samtools_output: $ERRNO";

  waitpid $pid, 0;

  return 1;
}

sub _version_info {
  my $self = shift;

  eval {
    my $samtools_version = npg_qc::autoqc::checks::check->aligner_version($self->samtools_cmd());
    if($samtools_version){
       $self->_result->set_info('Samtools', $samtools_version);
    }
    1;
  } or do {
    carp $EVAL_ERROR;
  };

  if($self->no_alignment()){
     return 1;
  }

  my $picard_version = $self->_get_picard_version();
  if($picard_version){
     $self->_result->set_info('Picard-tools', $picard_version);
  }

  return 1;
}

sub _get_picard_version {
  my $self = shift;

  my $picard_path = abs_path($self->jar_file());
  my ($picard_version) = $picard_path =~ m/ picard-tools- ( \d+ [.] \d+ )/mxs;
  return $picard_version;
}

no Moose;

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item MooseX::Getopt

=item Carp

=item English -no_match_vars

=item Readonly

=item Perl6::Slurp

=item File::Temp

=item IPC::Open3

=item Cwd qw(abs_path)

=item File::Basename

=item autodie qw(:all)

=item npg_qc::autoqc::results::bam_flagstats

=item npg_qc::autoqc::checks::check

=item npg_common::roles::log

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Guoying Qi E<lt>gq1@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2011 GRL, by Guoying Qi

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
