#########
# Author:        gq1
# Maintainer:    $Author: gq1 $
# Created:       2010-10-28
# Last Modified: $Date: 2010-11-19 12:08:06 +0000 (Fri, 19 Nov 2010) $
# Id:            $Id: SAM_Fastq_Check.pm 11838 2010-11-19 12:08:06Z gq1 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-43.0/lib/npg_common/sequence/SAM_Fastq_Check.pm $
#

package npg_common::sequence::SAM_Fastq_Check;

use strict;
use warnings;
use Moose;
use Carp;
use English qw(-no_match_vars);

with 'MooseX::Getopt';
with 'npg_common::roles::log';

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$Revision: 11838 $ =~ /(\d+)/smx; $r; };
## no critic (Documentation::RequirePodAtEnd)

=head1 NAME

npg_common::sequence::SAM_Fastq_Check

=head1 VERSION

$LastChangedRevision: 11838 $

=head1 SYNOPSIS

=head1 DESCRIPTION

This modules will read records from sam file or sam stream, check the bases and qualities are the same as the ones in fastq files

The order of reads in sam file should match the order in fastq files

=head1 SUBROUTINES/METHODS

=head2 sam

input sam file, the order of the rescords should be the original

if missing, stdin will be used
 
=cut

has 'sam'             => (isa             => 'Str',
                          is              => 'rw',
                          documentation   => 'Input sam to check, if missing stdin will be used',
                         );

=head2 fastq_files

one or two fastq file names in order 

=cut

has 'fastq_files'    => ( isa             => 'ArrayRef',
                          is              => 'rw',
                          required        => 1,
                          documentation   => 'fasta files in array',
                        );

=head2 process

the main method to call

=cut

sub process{
   my $self = shift;

   my $fastq_files = $self->fastq_files();
   my @fastq_fhs;

   if(! scalar @{$fastq_files}){
      croak 'No fastq files given to check';
   }

   for my $fastq (@{$fastq_files}){
      if( ! $fastq || (! -e $fastq) ){
         croak "The given fastq file $fastq not exists!";
      }
      open my $fastq_fh, q{<}, $fastq or croak "Can not open file $fastq"; ## no critic (InputOutput::RequireBriefOpen)      
      push  @fastq_fhs, $fastq_fh;
   }

   #check where to get sam file and pre-process sam file
   my $sam_file = $self->sam();

   my $sam_fh;
   my $sam_line;

   #sam from a file or stdin
   if($sam_file){
     open $sam_fh, q{<}, $sam_file or croak "Can not open file:$sam_file"; ## no critic (InputOutput::RequireBriefOpen)
   }else{
     $sam_fh =q{STDIN};
   }

   #jump header section of sam
   $sam_line = <$sam_fh>;
   while($sam_line && $sam_line =~ /^@/mxs){
         print $sam_line or croak 'Can not print';
         $sam_line = <$sam_fh>;
   }

   chomp $sam_line;
   while( $sam_line ){
      #checking sam reconds with fastq files here
      if( ! $self->_check_sam_line_with_fastq($sam_line, \@fastq_fhs) ){
         croak "bases or quality scores in $sam_line do not match fastq's";
      }

      #if bases and qualities are the same, output
      print "$sam_line\n" or croak "Cannot print $sam_line";
      $sam_line = <$sam_fh>;
      if($sam_line){
         chomp $sam_line;
      }
   }

   if( $sam_fh ){
     close $sam_fh or croak "Can not close file: $sam_file";
   }

   for my $fastq_fh (@fastq_fhs){
         close $fastq_fh or croak 'Can not close fastq file';
   }

   return 1;
}

=head2 _check_sam_line_with_fastq

check bases and qualities are the same as the ones in fastq files

=cut

sub _check_sam_line_with_fastq {
   my ($self, $sam_line, $fastq_fhs) = @_;

   if(!$sam_line){
     croak 'The given sam line is empty';
   }

   my @sam_fields = split /\t/mxs, $sam_line;

   if( scalar  @sam_fields < 11 ) {## no critic (ValuesAndExpressions::ProhibitMagicNumbers)
     croak "SAM format wrong: $sam_line";
   }

   my $id_in_sam = shift @sam_fields;
   my $flag = shift @sam_fields;
   my $first_read_sam = ($flag & 64) >> 6; ## no critic (ValuesAndExpressions::ProhibitMagicNumbers)
   my $second_read_sam = ($flag & 128) >> 7; ## no critic (ValuesAndExpressions::ProhibitMagicNumbers)
   my $strand_query = ( $flag & 16 ) >> 4; ## no critic (ValuesAndExpressions::ProhibitMagicNumbers)

   my $bases_in_sam = $sam_fields[7]; ## no critic (ValuesAndExpressions::ProhibitMagicNumbers)
   my $qual_in_sam  = $sam_fields[8]; ## no critic (ValuesAndExpressions::ProhibitMagicNumbers)

   if($strand_query){

      $bases_in_sam = reverse $bases_in_sam;
      $bases_in_sam =~ tr/atcgATCG/tagcTAGC/;
      $qual_in_sam = reverse $qual_in_sam;
   }

   my $current_fastq_fh;
   if ($first_read_sam){
       $current_fastq_fh = $fastq_fhs->[0];
   }elsif( $second_read_sam ) {
       $current_fastq_fh = $fastq_fhs->[1];
   }elsif( scalar @{$fastq_fhs} == 1 ) {
      $current_fastq_fh = $fastq_fhs->[0];
   }else{
      croak "Read number is not set in sam file: $sam_line";
   }

   my $id_in_fastq = <$current_fastq_fh>;
   chomp $id_in_fastq;
   my $bases_in_fastq = <$current_fastq_fh>;
   chomp $bases_in_fastq;
   <$current_fastq_fh>;
   my $qual_in_fastq = <$current_fastq_fh>;
   chomp $qual_in_fastq;

   if($id_in_fastq !~ /$id_in_sam/mxs){
      croak "The order of reads in sam and fastq file not match:  $id_in_fastq - $id_in_sam";
   }

   if( $bases_in_fastq ne $bases_in_sam  || $qual_in_fastq ne $qual_in_sam ){
      $self->log("Fastq: $bases_in_fastq $qual_in_fastq");
      $self->log("SAM  : $bases_in_sam $qual_in_sam");
      return 0;
   }

   return 1;
}

no Moose;

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item MooseX::Getopt

=item Carp

=item English -no_match_vars

=item Readonly

=item npg_common::roles::log

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Guoying Qi E<lt>gq1@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Guoying Qi

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
