# Author:        John O'Brien and Marina Gourtovaia
# Maintainer:    $Author: mg8 $
# Created:       26 January 2010
# Last Modified: $Date: 2010-01-13 15:07:25 +0000 (Wed, 13 Jan 2010) $
# Id:            $Id: reference.pm 7844 2010-01-13 15:07:25Z mg8 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-qc/branches/prerelease-21.0/lib/npg_qc/autoqc/align/reference.pm $
#

package npg_common::sequence::reference::base_count;

use strict;
use warnings;
use Moose;
use Moose::Util::TypeConstraints;
use MooseX::ClassAttribute;
use Carp;
use IO::File;
use Fatal qw( open close );
use Readonly;

use MooseX::Storage;

with Storage( 'traits' => ['OnlyWhenBuilt'],
              'format' => 'JSON',
              'io'     => 'File' );


our $VERSION    = do { my ($r) = q$Revision: 7844 $ =~ /(\d+)/smx; $r; };
## no critic (Documentation::RequirePodAtEnd ProhibitParensWithBuiltins)

=head1 NAME

npg_common::sequence::reference::base_count

=head1 VERSION

$Revision: 7844 $

=head1 SYNOPSIS

 my $bc =  npg_common::sequence::reference::base_count->new(
                                        reference_path => q[t/data/sequence/base_count/three_seq.ref],
                                        bin_size => 4,
                                        slide    => 1,
                                                           );
 $bc->run;
 my $summary = $bc->summary;
 foreach my $bin (@{$bc->bins}) {
   print "Sequence name: $bin->{sequence_name}\n";
   print "Base\tCount\n";
   foreach my $base (keys %{$bin->{counts}}) {
     print "$base\t$$bin->{counts}->{$base}\n";
   }
 }

=head1 DESCRIPTION

The start of any subsequence and the first base code after a gap triggers a new bin creation.
The index is reset at the end of each subsequence. The index reflects the character
count (including gap characters in the subsequence).
Gap characters (- and .) are treated as the end of the sequence, ie the longest bin is closed
and saved, and shorter bins are discarded.

If too much memory is required for small bins&large refs, the algorithm can be re-implemented with PDL.
Another improvement will be not to store seq names with each bin, but just indices into an array with seq names.x

=head1 SUBROUTINES/METHODS

=cut


Readonly::Scalar my $MAX_WINDOW_SIZE  => 9_007_199_254_740_992; # max int on a 32-bit machine
Readonly::Scalar my $SEQ_NAME_LINE_TAG  => 62;

Readonly::Array  my @IUPAC_BASE_CODES     => qw(A C G T M R W S Y K V H D B N);
Readonly::Scalar my $GAP_CHAR => q[_];
Readonly::Scalar my $GAP_CHAR_1 => q[.];
Readonly::Scalar my $N_EQUIVALENT => q[X];
Readonly::Scalar my $COUNTS_HASH  => q[counts];
Readonly::Scalar my $NA           => 1;


subtype 'PositiveInteger'  => as 'Int'
                           => where { $_ > 0; }
                           => message { "$_ must be a positive integer" };

=head2 output_base_codes

Class attribute. Returns a ref to a list of output base codes

=cut
class_has '_output_base_codes' => (isa             => 'ArrayRef',
                                   is              => 'ro',
                                   required        => 0,
                                   lazy_build      => 1,
                                   reader          => 'output_base_codes',
);

sub _build__output_base_codes {
    return \@IUPAC_BASE_CODES;
}


=head2 translation_table

Class attribute. Returns a ref to a hash with a translation table from ASCII characters
to base output codes.

=cut
class_has '_translation_table' => (isa             => 'HashRef',
                                   is              => 'ro',
                                   required        => 0,
                                   reader          => 'translation_table',
                                   lazy_build      => 1,
);


sub _build__translation_table {

    my $hash = {};
    my @codes = ();

    foreach my $code (@IUPAC_BASE_CODES) {
        $hash->{ord($code)}    = $code;
        $hash->{ord(lc $code)} = $code;
    }
    $hash->{ord($N_EQUIVALENT)}    = q[N];
    $hash->{ord(lc $N_EQUIVALENT)} = q[N];

    $hash->{ord($GAP_CHAR)}   = q[];
    $hash->{ord($GAP_CHAR_1)} = q[];

    return $hash;
}


=head2 default_bin_size

Class attribute. Returns default bin size.

=cut
class_has '_default_bin_size' => (isa             => 'Int',
                                  is              => 'ro',
                                  required        => 0,
                                  reader          => 'default_bin_size',
                                  default         => $MAX_WINDOW_SIZE,
);

=head2 reference_path

Reference sequence file path

=cut
has 'reference_path' => ( isa           => 'Str',
                          is            => 'ro',
                          required      => 1,
                          documentation => 'Path to a reference fasta file',
);


=head2 bin_size

Requested bin size. The default bin size is a very large number, bigger
than the length of any reference sequence.

=cut
has 'bin_size'       => ( isa           => 'PositiveInteger',
                          is            => 'ro',
                          required      => 0,
                          default       => $MAX_WINDOW_SIZE,
);


=head2 slide

The shift between the starts of bins. Does not have to be less that the bin_size.
If not supplied, the value of the bin_size attribute is assigned.

=cut
has 'slide'         => ( isa           => 'PositiveInteger',
                         is            => 'ro',
                         required      => 0,
                         writer        => '_set_slide',
);


=head2 bins

A private variable with a public reader. An array for storing references to hashes
with info about a window/bin.

=cut
has '_bins'        => ( isa           => 'ArrayRef',
                        is            => 'rw',
                        required      => 0,
                        reader        => 'bins',
                        default       => sub { [] },
);


=head2 summary

A private variable with a public reader. A reference to a hash with information
about a sequence as a whole.

=cut
has '_summary'     => ( isa           => 'HashRef',
                        is            => 'rw',
                        required      => 0,
                        reader        => 'summary',
                        lazy_build    => 1,
);

sub _build__summary {
    my $self = shift;
    my $sb = $self->_new_bin_map();
    $sb->{ref_length} = 0;
    return $sb;
}

=head2 BUILD

On object construction, performs some tasks required - should not be used as a public method

=cut

sub BUILD {
    my $self = shift;
    if (!$self->slide) {
        $self->_set_slide($self->bin_size);
    }
    return;
}


=head2 run

Performs the base count. The results can be accessed through the bin and summary methods.

=cut
sub run {

    my ($self) = @_;

    if (@{$self->bins} > 0) {
        return;
    }
    if ( !-r $self->reference_path) {
        croak $self->reference_path . q[does not exist or is not readable];
    }
    my $fh = IO::File->new( $self->reference_path(), q[<] );
    $self->_parse_reference($fh);
    $fh->close();

    return;
}


=head2 actual_bin_size

Computes and returns the actual size of the argument bin

=cut
sub actual_bin_size {
    my ($self, $bin) = @_;
    my $total = 0;
    foreach my $value (values %{$bin->{$COUNTS_HASH}}) {
        $total += $value;
    }
    return $total;
}


=head2 gc_content4bin

Computes GC content for a bin

=cut
sub gc_content4bin {
    my ($self, $bin) = @_;

    my $length = 0;
    foreach my $base ( keys %{$bin->{$COUNTS_HASH}}) {
        $length += $bin->{$COUNTS_HASH}->{$base};
    }

    # S is the IUPAC ambiguity code for "C or G"
    my @certain = qw/C G S/;
    my $gc_count = 0;
    foreach my $base (@certain) {
        if (exists $bin->{$COUNTS_HASH}->{$base}) {
            $gc_count += $bin->{$COUNTS_HASH}->{$base};
	}
    }

    my @ambigious = qw/R Y M K H B D V N/;
    my $ambiguities = 0;
    foreach my $base (@ambigious) {
        if (exists $bin->{$COUNTS_HASH}->{$base}) {
            $ambiguities += $bin->{$COUNTS_HASH}->{$base};
	}
    }

    my $min_gc = $gc_count / $length;
    my $max_gc = ( $gc_count + $ambiguities ) / $length;

    # We could consider counting half of these ambiguity codes.
    #my $kmry_count = () = $fragment =~ m/([KMRY])/gimsx;
    #$kmry_count = $kmry_count / 2;

    # Two thirds of these.
    #### no critic (ProhibitMagicNumbers)
    #my $bv_count = () = $fragment =~ m/([BV])/gimsx;
    #$bv_count = ( 2 * $bv_count ) / 3;

    # And one third of these.
    #my $dh_count = () = $fragment =~ m/([DH])/gimsx;
    #$dh_count = $dh_count / 3;

    ## use critic
    #my $guess_gc =
    #    ( $gc_count + $kmry_count + $bv_count + $dh_count ) / $length;

    #my $n_count = () = $fragment =~ m/(N)/gimsx;

    return $min_gc, $max_gc;
}


sub _new_bin_map {

    my ($self, $start_coord) = @_;

    my $map = {};
    $map->{counts} = {};
    if (defined $start_coord) {
        if ($start_coord < 0) {
            croak q[Start coordinate should be non-negative ingeger];
        }
        $map->{start_coord} = $start_coord;
    }

    return $map;
}


sub _process_bin {

    my ($self, $bin_map_list, $seq_name, $clear)  = @_;

    if (!$bin_map_list || !@{$bin_map_list}) { return; }

    my $bin_map = shift @{$bin_map_list};
    $bin_map->{sequence_name} = $seq_name;

    # Append the bin to the list of bins
    push @{$self->_bins}, $bin_map;

    if ($clear) {
        @{$bin_map_list} = ();
    }

    return;
}


sub _parse_reference {

    my ($self, $fh) = @_;

    ## no critic (ProhibitDeepNests)

    my $seq_name = q[];    # name of the current sequence
    my $win_size = $self->bin_size;
    my $slide = $self->slide;
    my $index = 0;         # current sequence position
    my $contig_length = 0; # length of a continious string of characters
                           # within a current sequence
    my $prev_ascii = $NA;   # ascii code for the previously read char
    my @bin_map_list = ();

    my $line = <$fh>;
    while ($line) {

        my $original_line = $line;
        # Remove all white spaces
        $line =~ s/\s//gmsx;
        if ($line) {

        # Convert a line to an array of ascii values for all characters in a string
        my @ascii_values = unpack('C*', $line);

        # Trap header lines.
        #carp $line_count;
        if ($ascii_values[0] == $SEQ_NAME_LINE_TAG) {
            # Process the biggest (the first)  bin of the previous sequence if any
            my $clear = 1;
            $self->_process_bin(\@bin_map_list, $seq_name, $clear);
            my $temp;
	    ($seq_name, $temp) = $original_line =~ m/^>(?:\s*)(\S+)/msx;
            #$seq_name = $1;
            $index         = 0; # reset the current index position
            $contig_length = 0; # reset the contig length

        } else {

            foreach my $ascii (@ascii_values) {

                if (!exists $self->translation_table->{$ascii}) {
                    croak "$line contains unexpected character $ascii";
		}

                my $char_out = $self->translation_table->{$ascii};

                # If this char is the gap char
                if (!$char_out) {
                    # If this is the fisrt gap after a contig'
                    # process the largest bin, discard all bins and reset the contig length
                    if ($self->translation_table->{$prev_ascii}) {
                        my $clear = 1;
                        $self->_process_bin(\@bin_map_list, $seq_name, $clear);
                        $contig_length = 0;
		    }
	        } else {
                    # This is not a gap!
                    # If the time is right, process the largest (the first) bin
                    if (@bin_map_list) {
                        my $diff = $contig_length - $win_size;
                        if( $diff >= 0 && $diff % $slide == 0) {
			    $self->_process_bin(\@bin_map_list, $seq_name);
                        }
		    }

                    # If the contig has been reset or a new bin is due to start,
                    # push a new bin onto a list of bins
                    if ($contig_length == 0 || $contig_length % $slide == 0) {
                         push @bin_map_list, $self->_new_bin_map($index);
		    }

                    # If $slide is bigger than $bin_width the array might be undefined or empty
                    foreach my $bin_map (@bin_map_list) {
                        $bin_map->{$COUNTS_HASH}->{$char_out}++;
		    }

                    # Always increase counts in the summary bin
                    $self->_summary->{$COUNTS_HASH}->{$char_out}++;

                    $contig_length++;
	        }

                $self->_summary->{ref_length}++;
	        $index++; # sequence position increment, gap or no gap
                $prev_ascii = $ascii; # save the previous character
            }
        }
        }

        $line = <$fh>; # read another line
    }

    # Process the largest of the remaining bins
    $self->_process_bin(\@bin_map_list, $seq_name);

    return;
}


__PACKAGE__->meta->make_immutable;
no Moose;

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item warnings

=item strict

=item Moose

=item Moose::Util::TypeConstraints

=item MooseX::ClassAttribute

=item Carp

=item IO::File

=item Fatal qw( open close )

=item Readonly

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

Author: Marina Gourtovaia E<lt>mg8@sanger.ac.ukE<gt>

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Marina Gourtovaia

This file is part of NPG.

NPG is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

=cut
