#########
# Author:        gq1
# Maintainer:    $Author: dj3 $
# Created:       2011-08-26
# Last Modified: $Date: 2012-02-23 07:24:13 +0000 (Thu, 23 Feb 2012) $
# Id:            $Id: 10-sequence-BAM_Alignment.t 15225 2012-02-23 07:24:13Z dj3 $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/useful_modules/tags/release-43.0/t/10-sequence-BAM_Alignment.t $
#

use strict;
use warnings;
use Carp;
use English qw{-no_match_vars};
use Test::More tests => 85;
use Test::Deep;
use Test::Exception;
use File::Temp qw(tempfile tempdir);
use Perl6::Slurp;

use_ok('npg_common::sequence::BAM_Alignment');

my $temp_dir = tempdir( CLEANUP => 0 );

{

  my $bam = npg_common::sequence::BAM_Alignment->new(
               { input             => 'input.bam',
                 is_paired_read    => 1,
                 non_consent_split => 1,
                 spiked_phix_split => 1,       
                 read_length       => 37,
                 output_prefix     => 'output',
                 aligner_cmd       => 't/bin/aligners/bwa/current/bwa',
                 samtools_cmd      => 't/bin/aligners/samtools/current/samtools',
                 temp_dir          => $temp_dir,
                 reference         => 'reference.fasta',
               });
               
  isa_ok( $bam, 'npg_common::sequence::BAM_Alignment', 'object test' );
  
  ok( ! $bam->no_alignment(), 'output bam will be created with alignment' );
  
  is($bam->output(), 'output.bam', 'correct ouput file name');
  is($bam->non_consented_output(),'output_human.bam', 'correct ouput file name for human');
  is($bam->spiked_phix_output(),'output_phix.bam', 'correct ouput file name for spiked phix');
  is($bam->_bam_output_fifo(), $temp_dir.q{/output_fifo.bam}, 'correct bam ouput fifo file name');
  is($bam->_human_bam_output_fifo(), $temp_dir.q{/human_output_fifo.bam}, 'correct human bam ouput fifo file name');
  
  like( $bam->aligner_cmd(), qr{t/bin/bwa$}, 'correct aligner command with absolute path' );
  like( $bam->samtools_cmd(), qr{t/bin/aligners/samtools/samtools-0.1.9/samtools$}, 'correct samtools command with absolute path' );

  is( $bam->bwa_aln_options(), '-q 15 -t 4', 'default bwa aln options');
  
  like( $bam->_generate_bwa_aln_command('reference.fasta', 'input.bam', 1),
        qr{t/bin/bwa aln -q 15 -t 4 reference.fasta -b1 input.bam > $temp_dir/1.sai$},
        'correct bwa aln command for read 1'
      );
  like( $bam->_generate_bwa_aln_command('human_reference.fasta', 'input.bam', 0, 1),
        qr{t/bin/bwa aln -q 15 -t 4 human_reference.fasta -b0 input.bam > $temp_dir/human0.sai$},
        'correct bwa aln human command for single read'
      );
      
  like ( $bam->_generate_bwa_sam_command('reference.fasta', ['input.bam','input.bam']),
        qr{t/bin/bwa sampe reference.fasta $temp_dir/1.sai $temp_dir/2.sai input.bam input.bam$},
         'correct bwa sampe command for paired read'
       );
  like ( $bam->_generate_bwa_sam_command('reference.fasta', ['input.bam'], 1),
         qr{t/bin/bwa samse reference.fasta $temp_dir/human0.sai input.bam$},
         'correct bwa samse command for single read and human split alignment'
       );
  $bam->bwa_sampe_options('-s');
  like ( $bam->_generate_bwa_sam_command('reference.fasta', ['input.bam','input.bam'], 1),
        qr{t/bin/bwa sampe -s reference.fasta $temp_dir/human1.sai $temp_dir/human2.sai input.bam input.bam$},
         'correct bwa sampe command for paired read and human split with -s option'
       );    
  is( scalar @{$bam->_bwa_aln_commands()}, 2, '2 bwa aln commands for paired-read input file');
  $bam->human_reference('human_reference.fasta');
  is( scalar @{$bam->_human_bwa_aln_commands()}, 2, '2 bwa aln commands for paired-read input file against human');

  like ( $bam->bwa_sam_command(),
         qr{t/bin/bwa sampe -s reference.fasta $temp_dir/1.sai $temp_dir/2.sai input.bam input.bam$},
         'correct bwa sampe command for paired read with -s option'
       );
  like ( $bam->_human_bwa_sam_command(),
         qr{t/bin/bwa sampe -s human_reference.fasta $temp_dir/human1.sai $temp_dir/human2.sai input.bam input.bam$},
         'correct human bwa sampe command for paired read with -s option'
       );
  
  my @bwa_sam_commands = sort keys %{$bam->_bwa_sam_commands()};
  is(scalar @bwa_sam_commands, 4, '4 bwa sam commands returned inclusing cat - tee commands');
  is($bwa_sam_commands[2], "cat input.bam | tee $temp_dir/fifo1.bam $temp_dir/human_fifo1.bam > /dev/null", 'first cat tee command');
  is($bwa_sam_commands[3], "cat input.bam | tee $temp_dir/fifo2.bam $temp_dir/human_fifo2.bam > /dev/null", 'second cat tee command');
  like ( $bam->bwa_sam_command(),
         qr{t/bin/bwa sampe -s reference.fasta $temp_dir/1.sai $temp_dir/2.sai $temp_dir/fifo1.bam $temp_dir/fifo2.bam$},
         'correct bwa sampe command for paired read with -s option using fifo'
       );
  like ( $bam->_human_bwa_sam_command(),
         qr{t/bin/bwa sampe -s human_reference.fasta $temp_dir/human1.sai $temp_dir/human2.sai $temp_dir/human_fifo1.bam $temp_dir/human_fifo2.bam$},
         'correct human bwa sampe command for paired read with -s option using fifo'
       );
  ok( -e "$temp_dir/fifo1.bam", 'first fifo created');
  ok( -e "$temp_dir/fifo2.bam", 'second fifo created');
  ok( -e "$temp_dir/human_fifo1.bam", 'first human fifo created');
  ok( -e "$temp_dir/human_fifo2.bam", 'second human fifo created');
  
  my $filter_cmd = qq{/software/bin/java -Xmx1000m -jar /software/solexa/bin/aligners/illumina2bam/illumina2bam/dist/AlignmentFilter.jar VALIDATION_STRINGENCY=SILENT CREATE_MD5_FILE=true IN=input.bam IN=${temp_dir}/human_output_fifo.bam IN=${temp_dir}/output_fifo.bam OUT=output_phix.bam OUT=output_human.bam OUT=output.bam};
  is( $bam->_bam_alignment_filter_command(), $filter_cmd, 'correct filter command');
  ok( -e "$temp_dir/output_fifo.bam", 'bam output fifo created before alignment filter');
  ok( -e "$temp_dir/human_output_fifo.bam", 'human bam output fifo created before alignment filter');
}

{
  local $ENV{LSB_MCPU_HOSTS} = q(sf-3-3-12 6);
  my $bam = npg_common::sequence::BAM_Alignment->new(
               { input             => 'input.bam',
                 is_paired_read    => 1,
                 non_consent_split => 1,
                 spiked_phix_split => 1,
                 read_length       => 37,
                 output_prefix     => 'output',
                 aligner_cmd       => 't/bin/aligners/bwa/current/bwa',
                 samtools_cmd      => 't/bin/aligners/samtools/current/samtools',
                 temp_dir          => $temp_dir,
                 reference         => 'reference.fasta',
               });


  is( $bam->bwa_aln_options(), '-q 15 -t 6', 'default (LSF aware) bwa aln options');

  like( $bam->_generate_bwa_aln_command('reference.fasta', 'input.bam', 1),
        qr{t/bin/bwa aln -q 15 -t 6 reference.fasta -b1 input.bam > $temp_dir/1.sai$},
        'correct (LSF aware) bwa aln command for read 1'
      );
}

{
  my $bam = npg_common::sequence::BAM_Alignment->new(
               { input             => 'input.bam',
                 is_paired_read    => 0,
                 non_consent_split => 0,
                 spiked_phix_split => 0,       
                 read_length       => 24,
                 output_prefix     => 'output',
                 temp_dir          => $temp_dir,
                 aligner_cmd       => 't/bin/aligners/bwa/current/bwa',
               });
               
  isa_ok($bam, 'npg_common::sequence::BAM_Alignment', 'object test');

  ok( $bam->no_alignment(), 'read too short no alignment for output' );
  
  $bam->clear_no_alignment();  
  $bam->read_length(76);
  ok( $bam->no_alignment(), 'no reference given, no alignment for output' );
  
  $bam->clear_no_alignment();
  $bam->reference('reference.fasta');
  $bam->human_reference('human_reference.fasta');
  is( scalar @{$bam->_bwa_aln_commands()}, 1, '1 bwa aln command for single-read input file');
  is( scalar @{$bam->_human_bwa_aln_commands()}, 1, '1 bwa aln command for single-read input file against human');

  like ( $bam->bwa_sam_command(),
         qr{t/bin/bwa samse reference.fasta $temp_dir/0.sai input.bam$},
         'correct bwa samse command for single read'
       );
  like ( $bam->_human_bwa_sam_command(),
         qr{t/bin/bwa samse human_reference.fasta $temp_dir/human0.sai input.bam$},
         'correct human bwa samse command for single read'
       );

  my @bwa_sam_commands = sort keys %{$bam->_bwa_sam_commands()};
  is(scalar @bwa_sam_commands, 1, 'only one bwa sam command returned without non consented splitting');
  ok(!$bam->_bam_alignment_filter_command(), 'no alignment filter required');

}

{
  my $bam = npg_common::sequence::BAM_Alignment->new(
               { input             => 'input.bam',
                 is_paired_read    => 0,
                 non_consent_split => 1,
                 spiked_phix_split => 1,       
                 output_prefix     => 'output',
                 temp_dir          => $temp_dir,
                 aligner_cmd       => 't/bin/aligners/bwa/current/bwa',
                 reference         => 'reference.fasta',
                 human_reference   => 'human_reference.fasta',
               });
               
  isa_ok($bam, 'npg_common::sequence::BAM_Alignment', 'object test');

  ok( !$bam->no_alignment(), 'default with alignment when reference given' );

  my @bwa_sam_commands = sort keys %{$bam->_bwa_sam_commands()};
  is(scalar @bwa_sam_commands, 3, '3 bwa sam command returned including 1 cat-tee command');

  ok( -e "$temp_dir/fifo0.bam", 'one fifo created');
  ok( -e "$temp_dir/human_fifo0.bam", 'one human fifo created');
  is($bwa_sam_commands[2], "cat input.bam | tee $temp_dir/fifo0.bam $temp_dir/human_fifo0.bam > /dev/null", 'one cat tee command');
  like ( $bam->bwa_sam_command(),
         qr{t/bin/bwa samse reference.fasta $temp_dir/0.sai $temp_dir/fifo0.bam$},
         'correct bwa samse command for single read using fifo'
       );
  like ( $bam->_human_bwa_sam_command(),
         qr{t/bin/bwa samse human_reference.fasta $temp_dir/human0.sai $temp_dir/human_fifo0.bam$},
         'correct human bwa samse command for single read using fifo'
       );

}

{
  my $bam = npg_common::sequence::BAM_Alignment->new(
               { input             => 't/data/sequence/6062_1#0.bam',
                 is_paired_read    => 1,
                 non_consent_split => 1,
                 spiked_phix_split => 1,       
                 output_prefix     => 'output',
                 temp_dir          => $temp_dir,
                 reference         => 'reference.fasta',
                 human_reference   => 'human_reference.fasta',
               });

  my $samtools_cmd = $bam->samtools_cmd();
  like( $samtools_cmd, qr{^/software/solexa/bin/aligners/samtools\S+}, 'default samtools command returned');
  
  SKIP: {
      skip 'because samtools not found', 12 if !-e $samtools_cmd;

      is( @{ $bam->_original_pg_lines()}, 8, 'correct number PG lines' );

      my $bwa_aln_pg_lines = $bam->_bwa_aln_pg_lines();
      like ($bwa_aln_pg_lines->[0], qr{^\@PG\tID\:bwa_aln\tPN:bwa\tPP:SplitBamByReadGroup}, 'correct first bwa aln pg');
      like ($bwa_aln_pg_lines->[1], qr{^\@PG\tID\:bwa_aln_1\tPN:bwa\tPP:bwa_aln}, 'correct second bwa aln pg');
      
      my $human_bwa_aln_pg_lines = $bam->_human_bwa_aln_pg_lines();
      like ($human_bwa_aln_pg_lines->[0], qr{^\@PG\tID\:bwa_aln\tPN:bwa\tPP:SplitBamByReadGroup}, 'correct first human bwa aln pg');
      like ($human_bwa_aln_pg_lines->[1], qr{^\@PG\tID\:bwa_aln_1\tPN:bwa\tPP:bwa_aln}, 'correct human second bwa aln pg');
   
      my $output_bam = $temp_dir.'/test_out.bam';
      my $pg_records = [];
      my $output_bam_fh;
      lives_ok { ($output_bam_fh, $pg_records) = 
                  $bam->generate_output_bam_fh($output_bam, $pg_records, $pg_records)} 'generate output bam stream';
      is( @{$pg_records}, 2, '2 PG returned');
      like ($pg_records->[0], qr{^\@PG\tID\:Picard_SamFormatConverter\tPN\:SamFormatConverter\tVN\:picard\-tools\-}, 'first program is Picard SamFormatConverter');
      like ($pg_records->[1], qr{^\@PG\tID\:samtools_fixmate\tPN:samtools}, 'second program is samtools fixmate');
      isa_ok( $output_bam_fh,  'GLOB', 'output file handle');
      close $output_bam_fh;

      my $command = 'cat t/data/sequence/5431_6#3.sam';
      $output_bam = "$temp_dir/test.bam";
      my $output_bam_md5 = $output_bam.q{.md5};
      lives_ok{ $bam->_bwa_sam_out_to_bam($command, $bwa_aln_pg_lines, $output_bam)} 'covert sam stream from a command to a bam with steps: samtools fixmate and merge input with bam out put';
      ok(-e $output_bam, 'bam generated');
  }

  my $error_message = 'id_run, position or tag_index need to be set to get lims information';
  
  ok(! $bam->_lims(), 'no id_run given, no lims object');
  ok(!$bam->_npg_api_run(), 'no id_run given, no npg api run object');

  $bam->clear_is_paired_read();
  ok(!$bam->is_paired_read(), 'croak when no id_run given');

  $bam->clear_spiked_phix_split();
  ok(! $bam->spiked_phix_split(), 'croak when no id_run given');

  $bam->clear_non_consent_split();
  ok(!$bam->non_consent_split(), 'croak when no id_run given');
}

my $cache = q[t/data/sequence];
my $cache_var_name = q[NPG_WEBSERVICE_CACHE_DIR];
local $ENV{$cache_var_name} = $cache;

{
  my $bam = npg_common::sequence::BAM_Alignment->new(
               { input             => 't/data/sequence/6062_1#0.bam',     
                 output_prefix     => 'output',
                 temp_dir          => $temp_dir,
                 id_run          => 5175,
                 position        => 1,
                 tag_index       => 1,
                 not_strip_bam_tag => 1,
               });
   isa_ok($bam->_lims(), 'st::api::lims', 'correct lims object');
   isa_ok($bam->_npg_api_run(), 'npg::api::run','correct npg api run object');
   is( $bam->_npg_api_run()->id_run(), 5175, 'correct id_run in npg api run object');

   ok($bam->is_paired_read(), 'run 5175 is paired read run');
   ok(! $bam->spiked_phix_split(), 'run 5175 lane 1 and plex 1 no need spiked phix split');
   ok(! $bam->non_consent_split(), 'run 5175 lane 1 and plex 1 no non-consented split');

   SKIP: {
      skip 'not in farm', 8 if !-d '/lustre/scratch103/sanger/srpipe/references';
      ok(!$bam->no_alignment(), 'run 5175 lane 1 and plex 1 neeeeds to be aligned');
   
      lives_ok { $bam->reference_info() } 'getting reference info';
      lives_ok { $bam->reference_info() } 'getting reference info';
      is ( $bam->reference(),  '/lustre/scratch103/sanger/srpipe/references/Bordetella_bronchiseptica/RB50/all/bwa/B_bronchiseptica_RB50.fasta', 'reference for run 5175 lane 1 plex 1');
      is( $bam->bwa_aln_options(),'-q 15 -t 4', 'correct bwa aln option');
      is( $bam->ref_dict(), '/lustre/scratch103/sanger/srpipe/references/Bordetella_bronchiseptica/RB50/all/picard/B_bronchiseptica_RB50.fasta.dict', 'correct ref dict');
      
      is( $bam->human_reference(), '/lustre/scratch103/sanger/srpipe/references/Homo_sapiens/1000Genomes/all/bwa/human_g1k_v37.fasta', 'correct human reference');
      is( $bam->human_ref_dict(), '/lustre/scratch103/sanger/srpipe/references/Homo_sapiens/1000Genomes/all/picard/human_g1k_v37.fasta.dict', 'correct human reference dict');
   }
   
   is( $bam->_generate_markduplicates_cmd("$temp_dir/input.bam"), "/software/solexa/bin/bam_mark_duplicate.pl --java_Xmx 6000 --input_bam $temp_dir/input.bam --replace_file --change_bam_header --sort_input --id_run 5175 --position 1 --tag_index 1 --output_bam $temp_dir/input_mk.bam --metrics_json $temp_dir/qc/input_bam_flagstats.json --not_strip_bam_tag", "correct markduplicates command");
   
   is( $bam->_generate_markduplicates_cmd("$temp_dir/input_phix.bam", 'phix'), "/software/solexa/bin/bam_mark_duplicate.pl --java_Xmx 6000 --input_bam $temp_dir/input_phix.bam --replace_file --change_bam_header --sort_input --id_run 5175 --position 1 --tag_index 1 --human_split phix --output_bam $temp_dir/input_phix_mk.bam --metrics_json $temp_dir/qc/input_phix_bam_flagstats.json --not_strip_bam_tag", "correct markduplicates command for phix part bam");
}


1;
__END__
